#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import logging
import sys
import unittest
from uuid import UUID

from onefuzztypes.enums import JobState
from onefuzztypes.models import Job, JobConfig
from onefuzztypes.primitives import Directory, File

from onefuzz.api import Onefuzz
from onefuzz.templates import JobHelper


class TestHelper(unittest.TestCase):
    def test_path_resolution(self) -> None:
        helper = JobHelper(
            Onefuzz(),
            logging.getLogger(),
            "a",
            "b",
            "c",
            False,
            target_exe=File("README.md"),
            job=Job(
                job_id=UUID("0" * 32),
                state=JobState.init,
                config=JobConfig(project="a", name="a", build="a", duration=1),
            ),
        )
        values = {
            (File("filename.txt"), None): "filename.txt",
            (File("dir/filename.txt"), None): "filename.txt",
            (File("./filename.txt"), None): "filename.txt",
            (File("./filename.txt"), Directory(".")): "filename.txt",
            (File("dir/filename.txt"), Directory("dir")): "filename.txt",
            (File("dir/filename.txt"), Directory("dir/")): "filename.txt",
            (File("dir/filename.txt"), Directory("./dir")): "filename.txt",
            (File("./dir/filename.txt"), Directory("./dir/")): "filename.txt",
        }

        expected = "filename.txt"
        if sys.platform == "linux":
            filename = File("/unused/filename.txt")
            values[(filename, None)] = expected
            values[(filename, Directory("/unused"))] = expected
            values[(filename, Directory("/unused/"))] = expected

        if sys.platform == "windows":
            for filename in [
                File("c:/unused/filename.txt"),
                File("c:\\unused/filename.txt"),
                File("c:\\unused\\filename.txt"),
            ]:
                values[(filename, None)] = expected
                values[(filename, Directory("c:/unused"))] = expected
                values[(filename, Directory("c:/unused/"))] = expected
                values[(filename, Directory("c:\\unused\\"))] = expected
                values[(filename, Directory("c:\\unused\\"))] = expected

        for (args, expected) in values.items():
            self.assertEqual(helper.target_exe_blob_name(*args), expected)

        with self.assertRaises(ValueError):
            helper.target_exe_blob_name(
                File("dir/filename.txt"), Directory("other_dir")
            )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import json
import logging
import os
import re
import subprocess  # nosec
import uuid
from shutil import which
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, TypeVar
from uuid import UUID

import pkg_resources
import semver
from onefuzztypes import enums, models, primitives, requests, responses
from pydantic import BaseModel
from six.moves import input  # workaround for static analysis

from .__version__ import __version__
from .backend import Backend, ContainerWrapper, wait
from .ssh import build_ssh_command, ssh_connect, temp_file

UUID_EXPANSION = TypeVar("UUID_EXPANSION", UUID, str)

DEFAULT = {
    "authority": "https://login.microsoftonline.com/72f988bf-86f1-41af-91ab-2d7cd011db47",
    "client_id": "72f1562a-8c0c-41ea-beb9-fa2b71c80134",
}

# This was generated randomly and should be preserved moving forwards
ONEFUZZ_GUID_NAMESPACE = uuid.UUID("27f25e3f-6544-4b69-b309-9b096c5a9cbc")

DEFAULT_LINUX_IMAGE = "Canonical:UbuntuServer:18.04-LTS:latest"
DEFAULT_WINDOWS_IMAGE = "MicrosoftWindowsDesktop:Windows-10:rs5-pro:latest"

REPRO_SSH_FORWARD = "1337:127.0.0.1:1337"

UUID_RE = r"^[a-f0-9]{8}-?[a-f0-9]{4}-?[a-f0-9]{4}-?[a-f0-9]{4}-?[a-f0-9]{12}\Z"


def is_uuid(value: str) -> bool:
    return bool(re.match(UUID_RE, value))


A = TypeVar("A", bound=BaseModel)


def wsl_path(path: str) -> str:
    if which("wslpath"):
        return subprocess.check_output(["wslpath", "-w", path]).decode().strip()
    return path


class Endpoint:
    endpoint: str

    def __init__(self, onefuzz: "Onefuzz"):
        self.onefuzz = onefuzz
        self.logger = onefuzz.logger

    def _req_model(
        self,
        method: str,
        model: Type[A],
        *,
        json_data: Any = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> A:
        response = self.onefuzz._backend.request(
            method, self.endpoint, json_data=json_data, params=params
        )
        return model.parse_obj(response)

    def _req_model_list(
        self,
        method: str,
        model: Type[A],
        *,
        json_data: Any = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> List[A]:
        response = self.onefuzz._backend.request(
            method, self.endpoint, json_data=json_data, params=params
        )
        return [model.parse_obj(x) for x in response]

    def _disambiguate(
        self,
        name: str,
        value: str,
        check: Callable[[str], bool],
        func: Callable[[], List[str]],
    ) -> str:
        if check(value):
            return value

        self.logger.debug("expanding %s: %s", name, value)

        values = [x for x in func() if x.startswith(value)]
        if len(values) == 1:
            return values[0]

        if len(values) > 1:
            if value in values:
                return value
            raise Exception(
                "%s expands to multiple values - %s: %s"
                % (name, value, ",".join(values))
            )

        raise Exception("%s does not expand to a value: %s" % (name, value))

    def _disambiguate_uuid(
        self,
        name: str,
        value: UUID_EXPANSION,
        func: Callable[[], List[str]],
    ) -> UUID:
        if isinstance(value, UUID):
            return value
        return UUID(self._disambiguate(name, value, is_uuid, func))


class Files(Endpoint):
    """ Interact with files within a container """

    endpoint = "files"

    def _get_client(self, container: str) -> ContainerWrapper:
        sas = self.onefuzz.containers.get(container).sas_url
        return ContainerWrapper(sas)

    def list(self, container: str) -> models.Files:
        """ Get a list of files in a container """
        self.logger.debug("listing files in container: %s", container)
        client = self._get_client(container)
        return models.Files(files=client.list_blobs())

    def delete(self, container: str, filename: str) -> None:
        """ delete a file from a container """
        self.logger.debug("deleting in container: %s:%s", container, filename)
        client = self._get_client(container)
        client.delete_blob(filename)

    def get(self, container: str, filename: str) -> bytes:
        """ get a file from a container """
        self.logger.debug("getting file from container: %s:%s", container, filename)
        client = self._get_client(container)
        return client.download_blob(filename)

    def upload_file(
        self, container: str, file_path: str, blob_name: Optional[str] = None
    ) -> None:
        """ uploads a file to a container """
        if not blob_name:
            # Default blob name to file basename. This means that the file data will be
            # written to the "root" of the container, if simulating a directory tree.
            blob_name = os.path.basename(file_path)

        self.logger.debug(
            "uploading file to container %s:%s (blob_name: %s)",
            container,
            file_path,
            blob_name,
        )

        client = self._get_client(container)
        client.upload_file(file_path, blob_name)

    def upload_dir(self, container: str, dir_path: str) -> None:
        """ uploads a directory to a container """

        self.logger.debug("uploading directory to container %s:%s", container, dir_path)

        client = self._get_client(container)
        client.upload_dir(dir_path)


class Versions(Endpoint):
    """ Onefuzz Instance """

    def check(self, exact: bool = False) -> str:
        """ Compare API and CLI versions for compatibility """
        versions = self.onefuzz.info.get().versions
        api_str = versions["onefuzz"].version
        cli_str = __version__
        if exact:
            result = semver.compare(api_str, cli_str) == 0
        else:
            api = semver.VersionInfo.parse(api_str)
            cli = semver.VersionInfo.parse(cli_str)
            result = (
                api.major > 0 and api.major == cli.major and api.minor >= cli.minor
            ) or (
                api.major == 0
                and api.major == cli.major
                and api.minor == cli.minor
                and api.patch >= cli.patch
            )
            if cli_str == "0.0.0" and not result:
                self.logger.warning(
                    "ignoring compatibility check as the CLI was installed "
                    "from git.  api: %s cli: %s",
                    api_str,
                    cli_str,
                )
                result = True

        if not result:
            raise Exception(
                "incompatible versions.  api: %s cli: %s" % (api_str, cli_str)
            )

        return "compatible"


class Info(Endpoint):
    """ Information about the OneFuzz instance """

    endpoint = "info"

    def get(self) -> responses.Info:
        """ Get information about the OneFuzz instance """
        self.logger.debug("getting info")
        return self._req_model("GET", responses.Info)


class Containers(Endpoint):
    """ Interact with Onefuzz containers """

    endpoint = "containers"

    def __init__(self, onefuzz: "Onefuzz"):
        super().__init__(onefuzz)
        self.files = Files(onefuzz)

    def get(self, name: str) -> responses.ContainerInfo:
        """ Get a fully qualified SAS URL for a container """
        self.logger.debug("get container: %s", name)
        return self._req_model("GET", responses.ContainerInfo, json_data={"name": name})

    def create(
        self, name: str, metadata: Optional[Dict[str, str]] = None
    ) -> responses.ContainerInfo:
        """ Create a storage container """
        self.logger.debug("create container: %s", name)
        return self._req_model(
            "POST",
            responses.ContainerInfo,
            json_data={"name": name, "metadata": metadata},
        )

    def delete(self, name: str) -> responses.BoolResult:
        """ Delete a storage container """
        self.logger.debug("delete container: %s", name)
        return self._req_model("DELETE", responses.BoolResult, json_data={"name": name})

    def list(self) -> List[responses.ContainerInfoBase]:
        """ Get a list of containers """
        self.logger.debug("list containers")
        return self._req_model_list("GET", responses.ContainerInfoBase)


class Repro(Endpoint):
    """ Interact with Reproduction VMs """

    endpoint = "repro_vms"

    def get(self, vm_id: UUID_EXPANSION) -> models.Repro:
        """ get information about a Reproduction VM """
        vm_id_expanded = self._disambiguate_uuid(
            "vm_id", vm_id, lambda: [str(x.vm_id) for x in self.list()]
        )

        self.logger.debug("get repro vm: %s", vm_id_expanded)
        return self._req_model("GET", models.Repro, json_data={"vm_id": vm_id_expanded})

    def create(self, container: str, path: str, duration: int = 24) -> models.Repro:
        """ Create a Reproduction VM from a Crash Report """
        self.logger.info(
            "creating repro vm: %s %s (%d hours)", container, path, duration
        )
        return self._req_model(
            "POST",
            models.Repro,
            json_data={"container": container, "path": path, "duration": duration},
        )

    def delete(self, vm_id: UUID_EXPANSION) -> models.Repro:
        """ Delete a Reproduction VM """
        vm_id_expanded = self._disambiguate_uuid(
            "vm_id", vm_id, lambda: [str(x.vm_id) for x in self.list()]
        )

        self.logger.debug("deleting repro vm: %s", vm_id_expanded)
        return self._req_model(
            "DELETE", models.Repro, json_data={"vm_id": vm_id_expanded}
        )

    def list(self) -> List[models.Repro]:
        """ List all VMs """
        self.logger.debug("listing repro vms")
        return self._req_model_list("GET", models.Repro, json_data={})

    def _dbg_linux(
        self, repro: models.Repro, debug_command: Optional[str]
    ) -> Optional[str]:
        """ Launch gdb with GDB script that includes 'target remote | ssh ...' """

        if (
            repro.auth is None
            or repro.ip is None
            or repro.state != enums.VmState.running
        ):
            raise Exception("vm setup failed: %s" % repro.state)

        with build_ssh_command(
            repro.ip, repro.auth.private_key, command="-T"
        ) as ssh_cmd:

            gdb_script = [
                "target remote | %s sudo /onefuzz/bin/repro-stdout.sh"
                % " ".join(ssh_cmd)
            ]

            if debug_command:
                gdb_script += [debug_command, "quit"]

            with temp_file("gdb.script", "\n".join(gdb_script)) as gdb_script_path:
                dbg = ["gdb", "--silent", "--command", gdb_script_path]

                if debug_command:
                    dbg += ["--batch"]

                    try:
                        return subprocess.run(
                            dbg, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
                        ).stdout.decode(errors="ignore")
                    except subprocess.CalledProcessError as err:
                        self.logger.error(
                            "debug failed: %s", err.output.decode(errors="ignore")
                        )
                        raise err
                else:
                    subprocess.call(dbg)
        return None

    def _dbg_windows(
        self, repro: models.Repro, debug_command: Optional[str]
    ) -> Optional[str]:
        """ Setup an SSH tunnel, then connect via CDB over SSH tunnel """

        if (
            repro.auth is None
            or repro.ip is None
            or repro.state != enums.VmState.running
        ):
            raise Exception("vm setup failed: %s" % repro.state)

        bind_all = which("wslpath") is not None and repro.os == enums.OS.windows
        proxy = "*:" + REPRO_SSH_FORWARD if bind_all else REPRO_SSH_FORWARD
        with ssh_connect(repro.ip, repro.auth.private_key, proxy=proxy):
            dbg = ["cdb.exe", "-remote", "tcp:port=1337,server=localhost"]
            if debug_command:
                dbg_script = [debug_command, "qq"]
                with temp_file("db.script", "\r\n".join(dbg_script)) as dbg_script_path:
                    dbg += ["-cf", wsl_path(dbg_script_path)]

                    logging.debug("launching: %s", dbg)
                    try:
                        return subprocess.run(
                            dbg, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
                        ).stdout.decode(errors="ignore")
                    except subprocess.CalledProcessError as err:
                        self.logger.error(
                            "debug failed: %s", err.output.decode(errors="ignore")
                        )
                        raise err
            else:
                logging.debug("launching: %s", dbg)
                subprocess.call(dbg)

        return None

    def connect(
        self,
        vm_id: UUID_EXPANSION,
        delete_after_use: bool = False,
        debug_command: Optional[str] = None,
    ) -> Optional[str]:
        """ Connect to an existing Reproduction VM """

        self.logger.info("connecting to reproduction VM: %s", vm_id)

        if which("ssh") is None:
            raise Exception("unable to find ssh")

        def missing_os() -> Tuple[bool, str, models.Repro]:
            repro = self.get(vm_id)
            return (
                repro.os is not None,
                "waiting for os determination",
                repro,
            )

        repro = wait(missing_os)

        if repro.os == enums.OS.windows:
            if which("cdb.exe") is None:
                raise Exception("unable to find cdb.exe")
        if repro.os == enums.OS.linux:
            if which("gdb") is None:
                raise Exception("unable to find gdb")

        def func() -> Tuple[bool, str, models.Repro]:
            repro = self.get(vm_id)
            state = repro.state
            return (
                repro.auth is not None
                and repro.ip is not None
                and state not in [enums.VmState.init, enums.VmState.extensions_launch],
                "launching reproducing vm.  current state: %s" % state,
                repro,
            )

        repro = wait(func)

        result: Optional[str] = None

        if repro.os == enums.OS.windows:
            result = self._dbg_windows(repro, debug_command)
        elif repro.os == enums.OS.linux:
            result = self._dbg_linux(repro, debug_command)
        else:
            raise NotImplementedError

        if delete_after_use:
            self.logger.debug("deleting vm %s", repro.vm_id)
            self.delete(repro.vm_id)

        return result

    def create_and_connect(
        self,
        container: str,
        path: str,
        duration: int = 24,
        delete_after_use: bool = False,
        debug_command: Optional[str] = None,
    ) -> Optional[str]:
        """ Create and connect to a Reproduction VM """
        repro = self.create(container, path, duration=duration)
        return self.connect(
            repro.vm_id, delete_after_use=delete_after_use, debug_command=debug_command
        )


class Notifications(Endpoint):
    """ Interact with models.Notifications """

    endpoint = "notifications"

    def create(
        self, container: str, config: models.NotificationConfig
    ) -> models.Notification:
        """ Create a notification based on a config file """

        config = requests.NotificationCreate(container=container, config=config.config)
        return self._req_model("POST", models.Notification, json_data=config)

    def create_teams(self, container: str, url: str) -> models.Notification:
        """ Create a Teams notification integration """

        self.logger.debug("create teams notification integration: %s", container)

        config = models.NotificationConfig(config=models.TeamsTemplate(url=url))
        return self.create(container, config)

    def create_ado(
        self,
        container: str,
        project: str,
        base_url: str,
        auth_token: str,
        work_item_type: str,
        unique_fields: List[str],
        comment: Optional[str] = None,
        fields: Optional[Dict[str, str]] = None,
        on_dup_increment: Optional[List[str]] = None,
        on_dup_comment: Optional[str] = None,
        on_dup_set_state: Optional[Dict[str, str]] = None,
        on_dup_fields: Optional[Dict[str, str]] = None,
    ) -> models.Notification:
        """ Create an Azure DevOps notification integration """

        self.logger.debug("create ado notification integration: %s", container)

        entry = models.NotificationConfig(
            config=models.ADOTemplate(
                base_url=base_url,
                auth_token=auth_token,
                project=project,
                type=work_item_type,
                comment=comment,
                unique_fields=unique_fields,
                ado_fields=fields,
                on_duplicate=models.ADODuplicateTemplate(
                    increment=on_dup_increment or [],
                    comment=on_dup_comment,
                    ado_fields=on_dup_fields or {},
                    set_state=on_dup_set_state or {},
                ),
            ),
        )
        return self.create(container, entry)

    def delete(self, notification_id: UUID_EXPANSION) -> models.Notification:
        """ Delete a notification integration """

        notification_id_expanded = self._disambiguate_uuid(
            "notification_id",
            notification_id,
            lambda: [str(x.notification_id) for x in self.list()],
        )

        self.logger.debug(
            "create notification integration: %s",
            notification_id_expanded,
        )
        return self._req_model(
            "DELETE",
            models.Notification,
            json_data={"notification_id": notification_id_expanded},
        )

    def list(self) -> List[models.Notification]:
        """ List notification integrations """

        self.logger.debug("listing notification integrations")
        return self._req_model_list("GET", models.Notification)


class Tasks(Endpoint):
    """ Interact with tasks """

    endpoint = "tasks"

    def delete(self, task_id: UUID_EXPANSION) -> models.Task:
        """ Stop an individual task """

        task_id_expanded = self._disambiguate_uuid(
            "task_id", task_id, lambda: [str(x.task_id) for x in self.list()]
        )

        self.logger.debug("delete task: %s", task_id_expanded)

        return self._req_model(
            "DELETE", models.Task, json_data={"task_id": task_id_expanded}
        )

    def get(self, task_id: UUID_EXPANSION) -> models.Task:
        """ Get information about a task """
        task_id_expanded = self._disambiguate_uuid(
            "task_id", task_id, lambda: [str(x.task_id) for x in self.list()]
        )

        self.logger.debug("get task: %s", task_id_expanded)

        return self._req_model(
            "GET", models.Task, json_data={"task_id": task_id_expanded}
        )

    def create(
        self,
        job_id: UUID_EXPANSION,
        task_type: enums.TaskType,
        target_exe: str,
        containers: List[Tuple[enums.ContainerType, primitives.Container]],
        *,
        pool_name: str,
        reboot_after_setup: bool = False,
        vm_count: int = 1,
        duration: int = 24,
        target_env: Optional[Dict[str, str]] = None,
        target_options: Optional[List[str]] = None,
        target_options_merge: bool = False,
        target_timeout: Optional[int] = None,
        rename_output: bool = False,
        check_asan_log: bool = False,
        check_debugger: bool = True,
        check_retry_count: Optional[int] = None,
        target_workers: Optional[int] = None,
        supervisor_exe: Optional[str] = None,
        supervisor_env: Optional[Dict[str, str]] = None,
        supervisor_options: Optional[List[str]] = None,
        supervisor_input_marker: Optional[str] = None,
        stats_file: Optional[str] = None,
        stats_format: Optional[enums.StatsFormat] = None,
        generator_exe: Optional[str] = None,
        generator_options: Optional[List[str]] = None,
        task_wait_for_files: Optional[enums.ContainerType] = None,
        analyzer_exe: Optional[str] = None,
        analyzer_options: Optional[List[str]] = None,
        analyzer_env: Optional[Dict[str, str]] = None,
        tags: Optional[Dict[str, str]] = None,
        prereq_tasks: Optional[List[UUID]] = None,
    ) -> models.Task:
        """ Create a task """
        self.logger.debug("creating task: %s", task_type)

        job_id_expanded = self._disambiguate_uuid(
            "job_id",
            job_id,
            lambda: [str(x.job_id) for x in self.onefuzz.jobs.list()],
        )

        if target_env is None:
            target_env = {}
        if tags is None:
            tags = {}
        if target_options is None:
            target_options = []
        if supervisor_options is None:
            supervisor_options = []
        if supervisor_env is None:
            supervisor_env = {}

        if prereq_tasks is None:
            prereq_tasks = []

        containers_submit = []
        all_containers = [x.name for x in self.onefuzz.containers.list()]
        for (container_type, container) in containers:
            if container not in all_containers:
                raise Exception("invalid container: %s" % container)
            containers_submit.append({"type": container_type, "name": container})

        data = {
            "job_id": job_id_expanded,
            "prereq_tasks": prereq_tasks,
            "task": {
                "type": task_type,
                "target_exe": target_exe,
                "duration": duration,
                "target_env": target_env,
                "target_options": target_options,
                "target_workers": target_workers,
                "target_options_merge": target_options_merge,
                "target_timeout": target_timeout,
                "rename_output": rename_output,
                "supervisor_exe": supervisor_exe,
                "supervisor_options": supervisor_options,
                "supervisor_env": supervisor_env,
                "supervisor_input_marker": supervisor_input_marker,
                "analyzer_exe": analyzer_exe,
                "analyzer_options": analyzer_options,
                "analyzer_env": analyzer_env,
                "stats_file": stats_file,
                "stats_format": stats_format,
                "generator_exe": generator_exe,
                "generator_options": generator_options,
                "wait_for_files": task_wait_for_files,
                "reboot_after_setup": reboot_after_setup,
                "check_asan_log": check_asan_log,
                "check_debugger": check_debugger,
                "check_retry_count": check_retry_count,
            },
            "pool": {"count": vm_count, "pool_name": pool_name},
            "containers": containers_submit,
            "tags": tags,
        }

        return self._req_model("POST", models.Task, json_data=data)

    def list(
        self,
        job_id: Optional[UUID_EXPANSION] = None,
        state: Optional[List[enums.TaskState]] = enums.TaskState.available(),
    ) -> List[models.Task]:
        """ Get information about all tasks """
        self.logger.debug("list tasks")
        job_id_expanded: Optional[UUID] = None

        if job_id is not None:
            job_id_expanded = self._disambiguate_uuid(
                "job_id",
                job_id,
                lambda: [str(x.job_id) for x in self.onefuzz.jobs.list()],
            )

        data = {"state": state, "job_id": job_id_expanded}

        return self._req_model_list("GET", models.Task, json_data=data)


class JobContainers(Endpoint):
    """ Interact with Containers used within tasks in a Job """

    endpoint = "jobs"

    def list(
        self,
        job_id: UUID_EXPANSION,
        container_type: Optional[
            enums.ContainerType
        ] = enums.ContainerType.unique_reports,
    ) -> Dict[str, List[str]]:
        """
        List the files for all of the containers of a given container type
        for the specified job
        """
        containers = set()
        tasks = self.onefuzz.tasks.list(job_id=job_id, state=[])
        for task in tasks:
            containers.update(
                set(x.name for x in task.config.containers if x.type == container_type)
            )

        results: Dict[str, List[str]] = {}
        for container in containers:
            results[container] = self.onefuzz.containers.files.list(container).files
        return results


class JobTasks(Endpoint):
    """ Interact with tasks within a job """

    endpoint = "jobs"

    def list(self, job_id: UUID_EXPANSION) -> List[models.Task]:
        """ List all of the tasks for a given job """
        return self.onefuzz.tasks.list(job_id=job_id, state=[])


class Jobs(Endpoint):
    """ Interact with Jobs """

    endpoint = "jobs"

    def __init__(self, onefuzz: "Onefuzz"):
        super().__init__(onefuzz)
        self.containers = JobContainers(onefuzz)
        self.tasks = JobTasks(onefuzz)

    def delete(self, job_id: UUID_EXPANSION) -> models.Job:

        """ Stop a job and all tasks that make up a job """
        job_id_expanded = self._disambiguate_uuid(
            "job_id", job_id, lambda: [str(x.job_id) for x in self.list()]
        )

        self.logger.debug("delete job: %s", job_id_expanded)
        return self._req_model(
            "DELETE", models.Job, json_data={"job_id": job_id_expanded}
        )

    def get(self, job_id: UUID_EXPANSION) -> models.Job:
        """ Get information about a specific job """
        job_id_expanded = self._disambiguate_uuid(
            "job_id", job_id, lambda: [str(x.job_id) for x in self.list()]
        )
        self.logger.debug("get job: %s", job_id_expanded)
        job = self._req_model("GET", models.Job, json_data={"job_id": job_id_expanded})
        # TODO
        # job["tasks"] = self.onefuzz.tasks.list(job_id=job["job_id"])
        return job

    def create(
        self, project: str, name: str, build: str, duration: int = 24
    ) -> models.Job:
        """ Create a job """
        self.logger.debug(
            "create job: project:%s name:%s build:%s", project, name, build
        )
        return self._req_model(
            "POST",
            models.Job,
            json_data={
                "project": project,
                "name": name,
                "build": build,
                "duration": duration,
            },
        )

    def list(
        self,
        job_state: Optional[List[enums.JobState]] = enums.JobState.available(),
    ) -> List[models.Job]:
        """ Get information about all jobs """
        self.logger.debug("list jobs")
        data = {"state": job_state}
        jobs = self._req_model_list("GET", models.Job, json_data=data)

        return jobs


class Pool(Endpoint):
    """ Interact with worker pools """

    endpoint = "pool"

    def create(
        self,
        name: str,
        os: enums.OS,
        client_id: Optional[UUID] = None,
        *,
        unmanaged: bool = False,
        arch: enums.Architecture = enums.Architecture.x86_64,
    ) -> models.Pool:
        self.logger.debug("create worker pool")
        managed = not unmanaged
        return self._req_model(
            "POST",
            models.Pool,
            json_data={
                "name": name,
                "os": os,
                "arch": arch,
                "managed": managed,
                "client_id": client_id,
            },
        )

    def get_config(self, pool_name: str) -> models.AgentConfig:
        """ Get the agent configuration for the pool  """

        pool = self._req_model("GET", models.Pool, json_data={"name": pool_name})

        if pool.config is None:
            raise Exception("Missing AgentConfig in response")

        config = pool.config
        config.client_credentials = models.ClientCredentials(  # nosec - bandit consider this a hard coded password
            client_id=pool.client_id,
            client_secret="<client secret>",
        )

        return config

    def shutdown(self, name: str, *, now: bool = False) -> models.Pool:
        expanded_name = self._disambiguate(
            "name", name, lambda x: False, lambda: [x.name for x in self.list()]
        )

        self.logger.debug("shutdown worker pool: %s (now: %s)", expanded_name, now)
        return self._req_model(
            "DELETE",
            models.Pool,
            json_data={"name": expanded_name, "now": now},
        )

    def get(self, name: str) -> models.Pool:
        self.logger.debug("get details on a specific pool")
        expanded_name = self._disambiguate(
            "name", name, lambda x: False, lambda: [x.name for x in self.list()]
        )

        return self._req_model("GET", models.Pool, json_data={"name": expanded_name})

    def list(
        self, *, state: Optional[List[enums.PoolState]] = None
    ) -> List[models.Pool]:
        self.logger.debug("list worker pools")
        return self._req_model_list("GET", models.Pool, json_data={"state": state})


class Node(Endpoint):
    """ Interact with nodes """

    endpoint = "node"

    def get(self, machine_id: UUID_EXPANSION) -> models.Node:
        self.logger.debug("get node: %s", machine_id)
        machine_id_expanded = self._disambiguate_uuid(
            "machine_id",
            machine_id,
            lambda: [str(x.machine_id) for x in self.list()],
        )

        return self._req_model(
            "GET", models.Node, json_data={"machine_id": machine_id_expanded}
        )

    def halt(self, machine_id: UUID_EXPANSION) -> responses.BoolResult:
        self.logger.debug("halt node: %s", machine_id)
        machine_id_expanded = self._disambiguate_uuid(
            "machine_id",
            machine_id,
            lambda: [str(x.machine_id) for x in self.list()],
        )

        return self._req_model(
            "DELETE",
            responses.BoolResult,
            json_data={"machine_id": machine_id_expanded},
        )

    def reimage(self, machine_id: UUID_EXPANSION) -> responses.BoolResult:
        self.logger.debug("reimage node: %s", machine_id)
        machine_id_expanded = self._disambiguate_uuid(
            "machine_id",
            machine_id,
            lambda: [str(x.machine_id) for x in self.list()],
        )

        return self._req_model(
            "PATCH", responses.BoolResult, json_data={"machine_id": machine_id_expanded}
        )

    def list(
        self,
        *,
        state: Optional[List[enums.NodeState]] = None,
        scaleset_id: Optional[UUID_EXPANSION] = None,
        pool_name: Optional[str] = None,
    ) -> List[models.Node]:
        self.logger.debug("list nodes")
        scaleset_id_expanded: Optional[UUID] = None

        if pool_name is not None:
            pool_name = self._disambiguate(
                "name",
                pool_name,
                lambda x: False,
                lambda: [x.name for x in self.onefuzz.pools.list()],
            )

        if scaleset_id is not None:
            scaleset_id_expanded = self._disambiguate_uuid(
                "scaleset_id",
                scaleset_id,
                lambda: [str(x.scaleset_id) for x in self.onefuzz.scalesets.list()],
            )

        return self._req_model_list(
            "GET",
            models.Node,
            json_data={
                "state": state,
                "scaleset_id": scaleset_id_expanded,
                "pool_name": pool_name,
            },
        )


class Scaleset(Endpoint):
    """ Interact with managed scaleset pools """

    endpoint = "scaleset"

    def _expand_scaleset_machine(
        self,
        scaleset_id: UUID_EXPANSION,
        machine_id: UUID_EXPANSION,
        *,
        include_auth: bool = False,
    ) -> Tuple[models.Scaleset, UUID]:
        scaleset = self.get(scaleset_id, include_auth=include_auth)

        if scaleset.nodes is None:
            raise Exception("no nodes defined in scaleset")
        nodes = scaleset.nodes

        machine_id_expanded = self._disambiguate_uuid(
            "machine_id", machine_id, lambda: [str(x.machine_id) for x in nodes]
        )
        return (scaleset, machine_id_expanded)

    def create(
        self,
        pool_name: str,
        size: int,
        *,
        image: Optional[str] = None,
        vm_sku: Optional[str] = "Standard_D2s_v3",
        region: Optional[str] = None,
        spot_instances: bool = False,
        tags: Optional[Dict[str, str]] = None,
    ) -> models.Scaleset:
        self.logger.debug("create scaleset")

        if tags is None:
            tags = {}

        if image is None:
            pool = self.onefuzz.pools.get(pool_name)
            if pool.os == enums.OS.linux:
                image = DEFAULT_LINUX_IMAGE
            elif pool.os == enums.OS.windows:
                image = DEFAULT_WINDOWS_IMAGE
            else:
                raise NotImplementedError

        return self._req_model(
            "POST",
            models.Scaleset,
            json_data={
                "pool_name": pool_name,
                "vm_sku": vm_sku,
                "image": image,
                "region": region,
                "size": size,
                "spot_instances": spot_instances,
                "tags": tags,
            },
        )

    def shutdown(
        self, scaleset_id: UUID_EXPANSION, *, now: bool = False
    ) -> models.Scaleset:
        scaleset_id_expanded = self._disambiguate_uuid(
            "scaleset_id",
            scaleset_id,
            lambda: [str(x.scaleset_id) for x in self.list()],
        )

        self.logger.debug("shutdown scaleset: %s (now: %s)", scaleset_id_expanded, now)
        return self._req_model(
            "DELETE",
            models.Scaleset,
            json_data={"scaleset_id": scaleset_id_expanded, "now": now},
        )

    def get(
        self, scaleset_id: UUID_EXPANSION, *, include_auth: bool = False
    ) -> models.Scaleset:
        self.logger.debug("get scaleset: %s", scaleset_id)
        scaleset_id_expanded = self._disambiguate_uuid(
            "scaleset_id",
            scaleset_id,
            lambda: [str(x.scaleset_id) for x in self.list()],
        )

        return self._req_model(
            "GET",
            models.Scaleset,
            json_data={
                "scaleset_id": scaleset_id_expanded,
                "include_auth": include_auth,
            },
        )

    def update(
        self, scaleset_id: UUID_EXPANSION, *, size: Optional[int] = None
    ) -> models.Scaleset:
        self.logger.debug("update scaleset: %s", scaleset_id)
        scaleset_id_expanded = self._disambiguate_uuid(
            "scaleset_id",
            scaleset_id,
            lambda: [str(x.scaleset_id) for x in self.list()],
        )

        return self._req_model(
            "PATCH",
            models.Scaleset,
            json_data={"scaleset_id": scaleset_id_expanded, "size": size},
        )

    def list(
        self,
        *,
        state: Optional[List[enums.ScalesetState]] = None,
    ) -> List[models.Scaleset]:
        self.logger.debug("list scalesets")
        return self._req_model_list("GET", models.Scaleset, json_data={"state": state})


class ScalesetProxy(Endpoint):
    """ Interact with Scaleset Proxies (NOTE: This API is unstable) """

    endpoint = "proxy"

    def delete(
        self,
        scaleset_id: UUID_EXPANSION,
        machine_id: UUID_EXPANSION,
        *,
        dst_port: Optional[int] = None,
    ) -> responses.BoolResult:
        """ Stop a proxy node """

        (
            scaleset,
            machine_id_expanded,
        ) = self.onefuzz.scalesets._expand_scaleset_machine(scaleset_id, machine_id)

        self.logger.debug(
            "delete proxy: %s:%d %d",
            scaleset.scaleset_id,
            machine_id_expanded,
            dst_port,
        )
        return self._req_model(
            "DELETE",
            responses.BoolResult,
            json_data={
                "scaleset_id": scaleset.scaleset_id,
                "machine_id": machine_id_expanded,
                "dst_port": dst_port,
            },
        )

    def reset(self, region: str) -> responses.BoolResult:
        """ Reset the proxy for an existing region """

        return self._req_model(
            "PATCH", responses.BoolResult, json_data={"region": region}
        )

    def get(
        self, scaleset_id: UUID_EXPANSION, machine_id: UUID_EXPANSION, dst_port: int
    ) -> responses.ProxyGetResult:
        """ Get information about a specific job """
        (
            scaleset,
            machine_id_expanded,
        ) = self.onefuzz.scalesets._expand_scaleset_machine(scaleset_id, machine_id)

        self.logger.debug(
            "get proxy: %s:%d:%d", scaleset.scaleset_id, machine_id_expanded, dst_port
        )
        proxy = self._req_model(
            "GET",
            responses.ProxyGetResult,
            json_data={
                "scaleset_id": scaleset.scaleset_id,
                "machine_id": machine_id_expanded,
                "dst_port": dst_port,
            },
        )
        return proxy

    def create(
        self,
        scaleset_id: UUID_EXPANSION,
        machine_id: UUID_EXPANSION,
        dst_port: int,
        *,
        duration: Optional[int] = 1,
    ) -> responses.ProxyGetResult:
        """ Create a proxy """
        (
            scaleset,
            machine_id_expanded,
        ) = self.onefuzz.scalesets._expand_scaleset_machine(scaleset_id, machine_id)

        self.logger.debug(
            "create proxy: %s:%s %d",
            scaleset.scaleset_id,
            machine_id_expanded,
            dst_port,
        )
        return self._req_model(
            "POST",
            responses.ProxyGetResult,
            json_data={
                "scaleset_id": scaleset.scaleset_id,
                "machine_id": machine_id_expanded,
                "dst_port": dst_port,
                "duration": duration,
            },
        )


class Command:
    def __init__(self, onefuzz: "Onefuzz", logger: logging.Logger):
        self.onefuzz = onefuzz
        self.logger = logger


class Utils(Command):
    def namespaced_guid(
        self,
        project: str,
        name: Optional[str] = None,
        build: Optional[str] = None,
        platform: Optional[str] = None,
    ) -> uuid.UUID:
        identifiers = [project]
        if name is not None:
            identifiers.append(name)
        if build is not None:
            identifiers.append(build)
        if platform is not None:
            identifiers.append(platform)
        return uuid.uuid5(ONEFUZZ_GUID_NAMESPACE, ":".join(identifiers))


class Onefuzz:
    def __init__(
        self, config_path: Optional[str] = None, token_path: Optional[str] = None
    ) -> None:
        self.logger = logging.getLogger("onefuzz")
        self._backend = Backend(
            config=DEFAULT, config_path=config_path, token_path=token_path
        )
        self.containers = Containers(self)
        self.repro = Repro(self)
        self.notifications = Notifications(self)
        self.tasks = Tasks(self)
        self.jobs = Jobs(self)
        self.versions = Versions(self)
        self.info = Info(self)
        self.scaleset_proxy = ScalesetProxy(self)
        self.pools = Pool(self)
        self.scalesets = Scaleset(self)
        self.nodes = Node(self)

        # these are externally developed cli modules
        self.template = Template(self, self.logger)
        self.debug = Debug(self, self.logger)
        self.status = Status(self, self.logger)
        self.utils = Utils(self, self.logger)

    def __setup__(self, endpoint: Optional[str] = None) -> None:
        if endpoint:
            self._backend.config["endpoint"] = endpoint

    def licenses(self) -> object:
        """ Return third-party licenses used by this package """
        stream = pkg_resources.resource_stream(__name__, "data/licenses.json")
        return json.load(stream)

    def logout(self) -> None:
        """ Logout of Onefuzz """
        self.logger.debug("logout")

        self._backend.logout()

    def login(self) -> str:
        """ Login to Onefuzz """

        # Rather than interacting MSAL directly, call a simple API which
        # actuates the login process
        self.info.get()
        return "succeeded"

    def config(
        self,
        endpoint: Optional[str] = None,
        authority: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
    ) -> Dict[str, str]:
        """ Configure onefuzz CLI """
        self.logger.debug("set config")

        if endpoint is not None:
            # The normal path for calling the API always uses the oauth2 workflow,
            # which the devicelogin can take upwards of 15 minutes to fail in
            # error cases.
            #
            # This check only happens on setting the configuration, as checking the
            # viability of the service on every call is prohibitively expensive.
            verify = self._backend.session.request("GET", endpoint)
            if verify.status_code != 401:
                self.logger.warning(
                    "This could be an invalid OneFuzz API endpoint: "
                    "Missing HTTP Authentication"
                )
            self._backend.config["endpoint"] = endpoint
        if authority is not None:
            self._backend.config["authority"] = authority
        if client_id is not None:
            self._backend.config["client_id"] = client_id
        if client_secret is not None:
            self._backend.config["client_secret"] = client_secret
        self._backend.app = None
        self._backend.save_config()

        data: Dict[str, str] = self._backend.config.copy()
        if "client_secret" in data:
            # replace existing secrets with "*** for user display
            data["client_secret"] = "***"  # nosec

        if not data["endpoint"]:
            self.logger.warning("endpoint not configured yet")

        return data

    def reset(self, everything: bool = False) -> None:
        """
        Resets onefuzz.  Stops all jobs, notifications, and repro VMs.
        Specifying 'everything' will delete all containers, pools, and managed
        scalesets.
        """
        message = ["Confirm stopping *all* running jobs"]
        if everything:
            message += ["AND deleting *ALL* fuzzing related data"]
        message += ["(specify y or n): "]

        answer: Optional[str] = None
        while answer not in ["y", "n"]:
            answer = input(" ".join(message)).strip()

        if answer == "n":
            self.logger.info("not stopping")
            return

        self.logger.info("stopping all interactions")

        for job in self.jobs.list():
            self.logger.info("stopping job %s", job.job_id)
            self.jobs.delete(job.job_id)

        for task in self.tasks.list():
            self.logger.info("stopping task %s", task.task_id)
            self.tasks.delete(task.task_id)

        for notification in self.notifications.list():
            self.logger.info("stopping notification %s", notification.notification_id)
            self.notifications.delete(notification.notification_id)

        for vm in self.repro.list():
            self.repro.delete(str(vm.vm_id))

        if everything:
            for pool in self.pools.list():
                self.logger.info("stopping pool: %s", pool.name)
                self.pools.shutdown(pool.name, now=True)

            for scaleset in self.scalesets.list():
                self.logger.info("stopping scaleset: %s", scaleset.scaleset_id)
                self.scalesets.shutdown(scaleset.scaleset_id, now=True)

            for container in self.containers.list():
                if (
                    container.metadata
                    and "container_type" in container.metadata
                    and container.metadata["container_type"]
                    in [
                        "analysis",
                        "coverage",
                        "crashes",
                        "inputs",
                        "no_repro",
                        "readonly_inputs",
                        "reports",
                        "setup",
                        "unique_reports",
                    ]
                ):
                    self.logger.info("removing container: %s", container.name)
                    self.containers.delete(container.name)


from .debug import Debug  # noqa: E402
from .status.cmd import Status  # noqa: E402
from .template import Template  # noqa: E402



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""
Simple CLI builder on top of a defined API (used for OneFuzz)
"""

import argparse
import inspect
import json
import logging
import os
import sys
import traceback
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from uuid import UUID

import jmespath
from docstring_parser import parse as parse_docstring
from msrest.serialization import Model
from onefuzztypes.primitives import Container, Directory, File
from pydantic import BaseModel, ValidationError

LOGGER = logging.getLogger("cli")

JMES_HELP = (
    "JMESPath query string. See http://jmespath.org/ "
    "for more information and examples."
)


def call_setup(api: Any, args: argparse.Namespace) -> None:
    setup = getattr(api, "__setup__", None)
    if setup is None:
        return

    myargs = {}
    for arg in inspect.getfullargspec(setup).args[1:]:
        if hasattr(args, arg):
            myargs[arg] = getattr(args, arg)

    for arg in inspect.getfullargspec(setup).kwonlyargs:
        if hasattr(args, arg):
            myargs[arg] = getattr(args, arg)

    setup(**myargs)


def call_func(func: Callable, args: argparse.Namespace) -> Any:
    myargs = {}
    for arg in inspect.getfullargspec(func).args[1:]:
        if hasattr(args, arg):
            myargs[arg] = getattr(args, arg)

    for arg in inspect.getfullargspec(func).kwonlyargs:
        if hasattr(args, arg):
            myargs[arg] = getattr(args, arg)
    return func(**myargs)


class AsDict(argparse.Action):
    def __call__(
        self,
        parser: argparse.ArgumentParser,  # noqa: F841 - unused args required by argparse
        namespace: argparse.Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: str = None,  # noqa: F841 - unused args required by argparse
    ) -> None:
        if values is None:
            return
        as_dict: Dict[str, str] = {k: v for k, v in (x.split("=", 1) for x in values)}
        setattr(namespace, self.dest, as_dict)


def arg_dir(arg: str) -> str:
    if not os.path.isdir(arg):
        raise argparse.ArgumentTypeError("not a directory: %s" % arg)
    return arg


def arg_file(arg: str) -> str:
    if not os.path.isfile(arg):
        raise argparse.ArgumentTypeError("not a file: %s" % arg)
    return arg


def is_optional(annotation: Any) -> bool:
    return is_a(annotation, Union, count=2) and (
        get_arg(annotation, 1) == type(None)  # noqa: E721
    )


def is_a(annotation: Any, origin: Any, count: Optional[int] = None) -> bool:
    return (
        (
            isinstance(origin, tuple)
            and getattr(annotation, "__origin__", None) in origin
        )
        or (
            not inspect.isclass(origin)
            and getattr(annotation, "__origin__", None) == origin
        )
    ) and (count is None or len(annotation.__args__) == count)


def is_dict(annotation: Any) -> bool:
    return is_a(annotation, dict, count=1)


def get_arg(annotation: Any, index: Optional[int] = None) -> Union[Any, List[Any]]:
    if index is None:
        return annotation.__args__
    else:
        return annotation.__args__[index]


def add_base(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-v", "--verbose", action="count", help="increase output verbosity", default=0
    )
    parser.add_argument(
        "--format", choices=["json", "raw"], default="json", help="output format"
    )
    parser.add_argument("--query", help=JMES_HELP)


def enum_help(entry: Type[Enum]) -> str:
    return "accepted %s: %s" % (entry.__name__, ", ".join(entry.__members__))


def tuple_help(entry: Any) -> str:
    doc = []
    for item in entry:
        if inspect.isclass(item) and issubclass(item, Enum):
            doc.append(
                "accepted %s: %s." % (item.__name__, ", ".join(item.__members__))
            )
    return " ".join(doc)


class Builder:
    def __init__(self, api_types: List[Any]):
        self.type_parsers: Dict[Any, Dict[str, Any]] = {
            str: {"type": str},
            int: {"type": int},
            UUID: {"type": UUID},
            Container: {"type": str},
            File: {"type": arg_file},
            Directory: {"type": arg_dir},
            Dict[str, str]: {"action": AsDict, "nargs": "+", "metavar": "key=val"},
        }
        self.api_types = tuple(api_types)
        self.top_level = argparse.ArgumentParser(add_help=False)
        add_base(self.top_level)

        self.main_parser = argparse.ArgumentParser()
        add_base(self.main_parser)

    def add_version(self, version: str) -> None:
        self.main_parser.add_argument(
            "--version",
            action="version",
            version="%(prog)s {version}".format(version=version),
        )

    def parse_api(self, api: Any) -> None:
        setup = getattr(api, "__setup__", None)
        if setup:
            self.parse_function(setup, self.main_parser)
        self.parse_nested_instances(self.main_parser, api)

    def get_help(self, obj: Any) -> str:
        return (parse_docstring(obj.__doc__).short_description or "").strip()

    def parse_function(self, func: Callable, parser: argparse.ArgumentParser) -> None:
        sig = inspect.signature(func)

        arg_docs = {}
        docs = parse_docstring(func.__doc__)
        for opt in docs.params:
            if opt.description:
                arg_docs[opt.arg_name] = opt.description
        for arg in sig.parameters:
            if arg == "self":
                continue
            help_doc = arg_docs.get(arg)
            args, kwargs = self.parse_param(arg, sig.parameters[arg], help_doc=help_doc)
            parser.add_argument(*args, **kwargs)

    def parse_param(
        self, name: str, param: inspect.Parameter, help_doc: Optional[str] = None
    ) -> Tuple[List[str], Dict[str, Any]]:
        """ Parse a single parameter """

        default = param.default
        annotation = param.annotation
        kwargs = self.parse_annotation(name, annotation, default, help_doc=help_doc)
        if not (
            isinstance(default, bool) or default in [None, inspect.Parameter.empty]
        ):
            if "help" in kwargs:
                kwargs["help"] += " (default: %(default)s)"
            else:
                kwargs["help"] = "(default: %(default)s)"
            kwargs["default"] = default

        optional = False

        if default is not inspect.Parameter.empty:
            optional = True

        if "optional" in kwargs:
            optional = True
            del kwargs["optional"]

        args = ["--" + name if optional else name]
        return args, kwargs

    def parse_annotation_class(
        self, name: str, annotation: Any, default: Any
    ) -> Optional[Dict[str, Any]]:
        if issubclass(annotation, Enum):
            result = {
                "type": annotation,
                "help": enum_help(annotation),
                "metavar": annotation.__name__,
            }
            return result

        if issubclass(annotation, bool) and default is False:
            return {
                "action": "store_true",
                "optional": True,
                "help": "(Default: False.  Sets value to True)",
            }

        if issubclass(annotation, bool) and default is True:
            return {
                "action": "store_false",
                "optional": True,
                "help": "(Default: True.  Sets value to False)",
            }

        if issubclass(annotation, BaseModel):

            def parse_model(data: str) -> object:
                if data.startswith("@"):
                    try:
                        with open(data[1:], "r") as handle:
                            data = handle.read()
                    except FileNotFoundError:
                        raise argparse.ArgumentTypeError("not a file: %s" % data[1:])

                try:
                    return annotation.parse_raw(data)
                except ValidationError as err:
                    raise argparse.ArgumentTypeError("parsing error\n" + str(err))

            parse_model.__name__ == annotation.__name__

            result = {
                "metavar": annotation.__name__,
                "help": "JSON for %s.  use @file to read from a file"
                % annotation.__name__,
                "type": parse_model,
            }
            return result

        return None

    def parse_annotation(
        self,
        name: str,
        annotation: Any,
        default: Any,
        help_doc: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Parse a single type annotation and get a signature appropriate
        for argparse.add_argument
        """

        result: Dict[str, Any] = {}
        if help_doc:
            result["help"] = help_doc

        if annotation in self.type_parsers:
            result.update(self.type_parsers[annotation].copy())
            return result

        if is_optional(annotation):
            result.update(self.parse_annotation(name, get_arg(annotation, 0), default))
            result["optional"] = True
            return result

        if is_a(annotation, (list, List), count=1):
            result.update(self.parse_annotation(name, get_arg(annotation, 0), default))
            result["nargs"] = "*"
            return result

        if is_a(annotation, (tuple, Tuple)):
            types = get_arg(annotation)

            def parse_tuple(data: str) -> Tuple[Any, ...]:
                split = data.split("=", len(types))
                if len(split) != len(types):
                    raise ValueError("invalid length: %s" % data)
                return tuple([x(y) for (x, y) in zip(types, split)])

            parse_tuple.__name__ = "=".join([x.__name__ for x in types])
            result.update(
                {
                    "metavar": parse_tuple.__name__,
                    "help": tuple_help(types),
                    "type": parse_tuple,
                }
            )
            return result

        if inspect.isclass(annotation):
            class_result = self.parse_annotation_class(name, annotation, default)
            if class_result is not None:
                result.update(class_result)
                return result

        # isinstance type signatures doesn't support TypeVar
        if hasattr(annotation, "__class__") and annotation.__class__ == TypeVar:

            def parse_typevar(data: str) -> object:
                for possible in annotation.__constraints__:
                    try:
                        return possible(data)
                    except ValueError:
                        pass
                raise argparse.ArgumentTypeError("Error parsing: %s" % data)

            result.update(
                {
                    "metavar": name,
                    "help": annotation.__name__,
                    "type": parse_typevar,
                }
            )
            return result

        raise Exception("unsupported annotation: %s - %s" % (name, annotation))

    def get_children(
        self, inst: Callable, is_callable: bool = False, is_typed: bool = False
    ) -> List[Tuple[str, Callable]]:
        entries = []
        for name in dir(inst):
            if name.startswith("_"):
                continue

            func = getattr(inst, name)
            if is_callable and not callable(func):
                continue

            if is_typed and not isinstance(func, self.api_types):
                continue

            entries.append((name, func))

        return entries

    def parse_instance(
        self, inst: Callable, subparser: argparse._SubParsersAction
    ) -> None:
        """ Expose every non-private callable in a class instance """
        for (name, func) in self.get_children(inst, is_callable=True):
            sub = subparser.add_parser(name, help=self.get_help(func))
            add_base(sub)
            self.parse_function(func, sub)
            sub.set_defaults(func=func)

    def parse_nested_instances(
        self, main_parser: argparse.ArgumentParser, inst: Callable, level: int = 0
    ) -> None:
        subparser = main_parser.add_subparsers(
            title="subcommands", dest="level_%d" % level
        )

        for (name, endpoint) in self.get_children(inst, is_typed=True):
            parser = subparser.add_parser(
                name, help=self.get_help(endpoint), parents=[self.top_level]
            )

            method_subparser = parser.add_subparsers(
                title="subcommands", dest="level_%d" % (level + 1)
            )

            for (nested_name, nested_endpoint) in self.get_children(
                endpoint, is_typed=True
            ):
                nested = method_subparser.add_parser(
                    nested_name,
                    help=self.get_help(nested_endpoint),
                    parents=[self.top_level],
                )
                self.parse_nested_instances(
                    nested,
                    nested_endpoint,
                    level=level + 2,
                )

            self.parse_instance(endpoint, method_subparser)

        self.parse_instance(inst, subparser)

    def parse_args(self) -> argparse.Namespace:
        return self.main_parser.parse_args()

    def print_nested_help(self, args: argparse.Namespace) -> None:
        level = 0
        parser = self.main_parser
        while True:
            if parser._subparsers is None:
                break
            if parser._subparsers._actions is None:
                break
            choices = parser._subparsers._actions[-1].choices
            value = getattr(args, "level_%d" % level)
            if value is None:
                parser.print_help()
                return

            if not choices:
                break
            if isinstance(choices, dict):
                parser = choices[value]
            else:
                return
            level += 1


def output(result: Any, output_format: str, expression: Optional[Any]) -> None:
    if isinstance(result, bytes):
        sys.stdout.buffer.write(result)
    else:
        if isinstance(result, list) and result and isinstance(result[0], BaseModel):
            # cycling through json resolves all of the nested BaseModel objects
            result = [json.loads(x.json(exclude_none=True)) for x in result]
        if isinstance(result, BaseModel):
            # cycling through json resolves all of the nested BaseModel objects
            result = json.loads(result.json(exclude_none=True))
        if isinstance(result, Model):
            result = result.as_dict()
        if expression is not None:
            result = expression.search(result)
        if result is not None:
            if output_format == "json":
                if isinstance(result, UUID):
                    result = str(result)
                result = json.dumps(result, indent=4, sort_keys=True)
            print(result, flush=True)


def execute_api(api: Any, api_types: List[Any], version: str) -> int:
    builder = Builder(api_types)
    builder.add_version(version)
    builder.parse_api(api)
    args = builder.parse_args()

    if args.verbose == 0:
        logging.basicConfig(level=logging.WARNING)
        api.logger.setLevel(logging.INFO)
    elif args.verbose == 1:
        logging.basicConfig(level=logging.WARNING)
        api.logger.setLevel(logging.INFO)
        logging.getLogger("nsv-backend").setLevel(logging.DEBUG)
    elif args.verbose == 2:
        logging.basicConfig(level=logging.INFO)
        api.logger.setLevel(logging.DEBUG)
        logging.getLogger("nsv-backend").setLevel(logging.DEBUG)
    elif args.verbose >= 3:
        logging.basicConfig(level=logging.DEBUG)
        api.logger.setLevel(logging.DEBUG)

    if not hasattr(args, "func"):
        LOGGER.error("no command specified")
        builder.print_nested_help(args)
        return 1

    if args.query:
        try:
            expression = jmespath.compile(args.query)
        except jmespath.exceptions.ParseError as err:
            LOGGER.error("unable to parse query: %s", err)
            return 1
    else:
        expression = None

    call_setup(api, args)

    try:
        result = call_func(args.func, args)
    except Exception as err:
        if args.verbose > 0:
            entry = traceback.format_exc()
            for x in entry.split("\n"):
                LOGGER.error("traceback: %s", x)
        LOGGER.error("command failed: %s", " ".join([str(x) for x in err.args]))
        return 1

    output(result, args.format, expression)

    return 0



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from datetime import datetime
from typing import Any, List, Sequence, Tuple

from asciimatics.event import KeyboardEvent
from asciimatics.exceptions import ResizeScreenError, StopApplication
from asciimatics.scene import Scene
from asciimatics.screen import Screen
from asciimatics.widgets import (
    Divider,
    Frame,
    Label,
    Layout,
    MultiColumnListBox,
    TextBox,
    Widget,
)

from .cache import TopCache, fmt


def now() -> datetime:
    return datetime.now()


# border + title + quit msg
BASE_LINES = 5

# divider + name + header
EXTRA_LINES_PER_WIDGET = 3


class TopView(Frame):
    def __init__(self, screen: Any, cache: TopCache, show_details: bool):

        super(TopView, self).__init__(
            screen, screen.height, screen.width, has_border=True, can_scroll=False
        )
        self.cache = cache
        self.show_details = show_details
        self.set_theme("monochrome")
        self.palette["title"] = (
            Screen.COLOUR_BLACK,
            Screen.A_NORMAL,
            Screen.COLOUR_WHITE,
        )

        self.job_count = len(self.cache.jobs)
        self.pool_count = len(self.cache.pools)

        max_widget_height = (
            int((screen.height - BASE_LINES) / 3) - EXTRA_LINES_PER_WIDGET
        )

        layout = Layout([1], fill_frame=True)
        self.add_layout(layout)

        self.onefuzz_reversed = {
            "pools": True,
            "jobs": True,
            "tasks": True,
            "messages": True,
        }

        dimensions = {
            "pools": {
                "height": min(self.pool_count + 1, max_widget_height),
                "setup": self.cache.POOL_FIELDS,
            },
            "jobs": {
                "height": min(self.job_count + 1, max_widget_height),
                "setup": self.cache.JOB_FIELDS,
            },
            "tasks": {"height": Widget.FILL_FRAME, "setup": self.cache.TASK_FIELDS},
            "messages": {
                "height": min(10, max_widget_height),
                "setup": [["Updated", "Type", "Message"], [1, 1, "100%"]],
            },
            "status": {"height": 1, "setup": [None, [1, "100%"]]},
        }

        for name in ["status", "pools", "jobs", "tasks", "messages"]:
            if name == "messages" and not self.show_details:
                continue

            if dimensions[name]["setup"][0]:
                title = TextBox(1, as_string=True, name=name + "_title")
                title.disabled = True
                title.custom_colour = "label"
                layout.add_widget(title)

            widget = MultiColumnListBox(
                dimensions[name]["height"],
                dimensions[name]["setup"][1],
                [],
                titles=dimensions[name]["setup"][0],
                name=name,
                add_scroll_bar=bool(dimensions[name]["setup"][0]),
            )
            if not dimensions[name]["setup"][0]:
                widget.disabled = True

            layout.add_widget(widget)
            layout.add_widget(Divider())

        layout.add_widget(Label("Press `q` to quit or `r` to reorder."))
        self.fix()

    def process_event(self, event: Any) -> Any:
        result = super(TopView, self).process_event(event)
        if isinstance(event, KeyboardEvent):
            # quit
            if event.key_code in [ord("q"), Screen.ctrl("c")]:
                raise StopApplication("")

            # toggle sort for current widget
            if event.key_code in [ord("r")]:
                if self.focussed_widget.name in self.onefuzz_reversed:
                    self.onefuzz_reversed[
                        self.focussed_widget.name
                    ] = not self.onefuzz_reversed[self.focussed_widget.name]

        return result

    def to_display(self, data: Sequence, reverse: bool) -> List[Tuple[str, int]]:
        data = sorted(data, reverse=reverse)
        data = [fmt(x) for x in data]
        return [(x, y) for (y, x) in enumerate(data)]

    def auto_resize(self, name: str) -> None:
        """ recompute widget width based on max length of all of the values """
        widget = self.find_widget(name)
        for column in range(len(widget._columns) - 1):

            sizes = [len(x[0][column]) + 1 for x in widget.options]
            if widget._titles:
                sizes.append(len(widget._titles[column]) + 1)
            widget._columns[column] = max(sizes)

    def render_base(self, name: str, data: Any) -> None:
        reverse = self.onefuzz_reversed.get(name, False)
        self.find_widget(name).options = self.to_display(data, reverse)
        self.auto_resize(name)
        title = self.find_widget(name + "_title")
        if title:
            title.value = "%s: %s" % (name.title(), fmt(len(data)))

    def update(self, frame_no: int) -> Any:
        if len(self.cache.jobs) != self.job_count:
            raise ResizeScreenError("resizing because of a differing job count")
        self.render_base("status", [[now(), "| " + self.cache.endpoint]])
        self.render_base("pools", self.cache.render_pools())
        self.render_base("jobs", self.cache.render_jobs())
        self.render_base("tasks", self.cache.render_tasks())

        if self.show_details:
            self.render_base("messages", self.cache.messages)

        super(TopView, self).update(frame_no)

    @property
    def frame_update_count(self) -> int:
        return 25


def render(data: TopCache, show_details: bool) -> None:
    while True:
        try:
            Screen.wrapper(
                lambda screen, data_ref, show_details: screen.play(
                    [Scene([TopView(screen, data_ref, show_details)], -1)],
                    stop_on_resize=True,
                ),
                catch_interrupt=True,
                arguments=[data, show_details],
            )
            return
        except ResizeScreenError:
            pass



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import logging
import os
import shutil
import subprocess  # nosec
import tempfile
from typing import Any, List, Optional, Tuple
from urllib.parse import urlparse
from uuid import UUID

from onefuzztypes.enums import ContainerType, TaskType
from onefuzztypes.models import BlobRef, NodeAssignment, Report, Task
from onefuzztypes.primitives import Directory

from onefuzz.api import UUID_EXPANSION, Command

from .backend import wait
from .rdp import rdp_connect
from .ssh import ssh_connect

EMPTY_SHA256 = "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"
ZERO_SHA256 = "0" * len(EMPTY_SHA256)


class DebugRepro(Command):
    """ Debug repro instances """

    def _disambiguate(self, vm_id: UUID_EXPANSION) -> str:
        return str(
            self.onefuzz.repro._disambiguate_uuid(
                "vm_id",
                vm_id,
                lambda: [str(x.vm_id) for x in self.onefuzz.repro.list()],
            )
        )

    def _info(self) -> Tuple[str, str]:
        info = self.onefuzz.info.get()
        return info.resource_group, info.subscription

    def ssh(self, vm_id: str) -> None:
        vm_id = self._disambiguate(vm_id)
        repro = self.onefuzz.repro.get(vm_id)
        if repro.ip is None:
            raise Exception("missing IP: %s" % repro)
        if repro.auth is None:
            raise Exception("missing Auth: %s" % repro)

        with ssh_connect(repro.ip, repro.auth.private_key, call=True):
            pass

    def rdp(self, vm_id: str) -> None:
        vm_id = self._disambiguate(vm_id)
        repro = self.onefuzz.repro.get(vm_id)
        if repro.ip is None:
            raise Exception("missing IP: %s" % repro)
        if repro.auth is None:
            raise Exception("missing Auth: %s" % repro)

        RDP_PORT = 3389
        with rdp_connect(repro.ip, repro.auth.password, port=RDP_PORT):
            return


class DebugScaleset(Command):
    """ Debug tasks """

    def _get_proxy_setup(
        self, scaleset_id: UUID, machine_id: UUID, port: int, duration: Optional[int]
    ) -> Tuple[bool, str, Optional[Tuple[str, int]]]:
        proxy = self.onefuzz.scaleset_proxy.create(
            scaleset_id, machine_id, port, duration=duration
        )
        if proxy.ip is None:
            return (False, "waiting on proxy", None)

        return (True, "waiting on proxy", (proxy.ip, proxy.forward.src_port))

    def rdp(
        self,
        scaleset_id: UUID_EXPANSION,
        machine_id: UUID_EXPANSION,
        duration: Optional[int] = 1,
    ) -> None:
        (
            scaleset,
            machine_id_expanded,
        ) = self.onefuzz.scalesets._expand_scaleset_machine(
            scaleset_id, machine_id, include_auth=True
        )

        RDP_PORT = 3389
        setup = wait(
            lambda: self._get_proxy_setup(
                scaleset.scaleset_id, machine_id_expanded, RDP_PORT, duration
            )
        )
        if setup is None:
            raise Exception("no proxy for RDP port configured")

        if scaleset.auth is None:
            raise Exception("auth is not available for scaleset")

        ip, port = setup
        with rdp_connect(ip, scaleset.auth.password, port=port):
            return

    def ssh(
        self,
        scaleset_id: UUID_EXPANSION,
        machine_id: UUID_EXPANSION,
        duration: Optional[int] = 1,
        command: Optional[str] = None,
    ) -> None:
        (
            scaleset,
            machine_id_expanded,
        ) = self.onefuzz.scalesets._expand_scaleset_machine(
            scaleset_id, machine_id, include_auth=True
        )

        SSH_PORT = 22
        setup = wait(
            lambda: self._get_proxy_setup(
                scaleset.scaleset_id, machine_id_expanded, SSH_PORT, duration
            )
        )
        if setup is None:
            raise Exception("no proxy for SSH port configured")

        ip, port = setup

        if scaleset.auth is None:
            raise Exception("auth is not available for scaleset")

        with ssh_connect(
            ip, scaleset.auth.private_key, port=port, call=True, command=command
        ):
            return


class DebugTask(Command):
    """ Debug a specific job """

    def list_nodes(self, task_id: UUID_EXPANSION) -> Optional[List[NodeAssignment]]:
        task = self.onefuzz.tasks.get(task_id)
        return task.nodes

    def _get_node(
        self, task_id: UUID_EXPANSION, node_id: Optional[UUID]
    ) -> Tuple[UUID, UUID]:
        nodes = self.list_nodes(task_id)
        if not nodes:
            raise Exception("task is not currently executing on nodes")

        if node_id is not None:
            for node in nodes:
                if node.node_id == node_id and node.scaleset_id is not None:
                    return (node.scaleset_id, node.node_id)
            raise Exception("unable to find scaleset with node_id")

        for node in nodes:
            if node.scaleset_id:
                return (node.scaleset_id, node.node_id)

        raise Exception("unable to find scaleset node running on task")

    def ssh(
        self,
        task_id: UUID_EXPANSION,
        *,
        node_id: Optional[UUID] = None,
        duration: Optional[int] = 1,
    ) -> None:
        scaleset_id, node_id = self._get_node(task_id, node_id)
        return self.onefuzz.debug.scalesets.ssh(scaleset_id, node_id, duration=duration)

    def rdp(
        self,
        task_id: UUID_EXPANSION,
        *,
        node_id: Optional[UUID] = None,
        duration: Optional[int] = 1,
    ) -> None:
        scaleset_id, node_id = self._get_node(task_id, node_id)
        return self.onefuzz.debug.scalesets.rdp(scaleset_id, node_id, duration=duration)


class DebugJobTask(Command):
    """ Debug a task for a specific job """

    def _get_task(self, job_id: UUID_EXPANSION, task_type: TaskType) -> UUID:
        for task in self.onefuzz.tasks.list(job_id=job_id):
            if task.config.task.type == task_type:
                return task.task_id

        raise Exception(
            "unable to find task type %s for job:%s" % (task_type.name, job_id)
        )

    def ssh(
        self,
        job_id: UUID_EXPANSION,
        task_type: TaskType,
        *,
        duration: Optional[int] = 1,
    ) -> None:
        """ SSH into the first node running the specified task type in the job """
        return self.onefuzz.debug.task.ssh(
            self._get_task(job_id, task_type), duration=duration
        )

    def rdp(
        self,
        job_id: UUID_EXPANSION,
        task_type: TaskType,
        *,
        duration: Optional[int] = 1,
    ) -> None:
        """ RDP into the first node running the specified task type in the job """
        return self.onefuzz.debug.task.rdp(
            self._get_task(job_id, task_type), duration=duration
        )


class DebugJob(Command):
    """ Debug a specific Job """

    def __init__(self, onefuzz: Any, logger: logging.Logger):
        super().__init__(onefuzz, logger)
        self.task = DebugJobTask(onefuzz, logger)

    def download_files(self, job_id: UUID_EXPANSION, output: Directory) -> None:
        """ Download the containers by container type for each task in the specified job """

        azcopy = os.environ.get("AZCOPY") or shutil.which("azcopy")
        if not azcopy:
            raise Exception(
                "unable to find 'azcopy' in path or AZCOPY environment variable"
            )

        to_download = {}
        tasks = self.onefuzz.tasks.list(job_id=job_id, state=None)
        if not tasks:
            raise Exception("no tasks with job_id:%s" % job_id)

        for task in tasks:
            for container in task.config.containers:
                info = self.onefuzz.containers.get(container.name)
                name = os.path.join(container.type.name, container.name)
                to_download[name] = info.sas_url

        for name in to_download:
            outdir = os.path.join(output, name)
            if not os.path.exists(outdir):
                os.makedirs(outdir)
            self.logger.info("downloading: %s", name)
            subprocess.check_output([azcopy, "sync", to_download[name], outdir])


class DebugNotification(Command):
    """ Debug notification integrations """

    def _get_container(
        self, task: Task, container_type: ContainerType
    ) -> Optional[str]:
        for container in task.config.containers:
            if container.type == container_type:
                return container.name
        return None

    def _get_storage_account(self, container_name: str) -> str:
        sas_url = self.onefuzz.containers.get(container_name).sas_url
        _, netloc, _, _, _, _ = urlparse(sas_url)
        return netloc.split(".")[0]

    def job(
        self,
        job_id: str,
        *,
        report_container_type: ContainerType = ContainerType.unique_reports,
        crash_name: str = "fake-crash-sample",
    ) -> None:
        """ Inject a report into the first crash reporting task in the specified job """

        tasks = self.onefuzz.tasks.list(job_id=job_id, state=[])
        for task in tasks:
            if task.config.task.type in [
                TaskType.libfuzzer_crash_report,
                TaskType.generic_crash_report,
            ]:
                self.task(
                    str(task.task_id),
                    report_container_type=report_container_type,
                    crash_name=crash_name,
                )
                return

        raise Exception("no crash reporting tasks configured")

    def task(
        self,
        task_id: str,
        *,
        report_container_type: ContainerType = ContainerType.unique_reports,
        crash_name: str = "fake-crash-sample",
    ) -> None:
        """ Inject a report into the specified crash reporting task """

        task = self.onefuzz.tasks.get(task_id)
        crashes = self._get_container(task, ContainerType.crashes)
        reports = self._get_container(task, report_container_type)

        if crashes is None:
            raise Exception("task does not have a crashes container")

        if reports is None:
            raise Exception(
                "task does not have a %s container", report_container_type.name
            )

        with tempfile.TemporaryDirectory() as tempdir:
            file_path = os.path.join(tempdir, crash_name)
            with open(file_path, "w") as handle:
                handle.write("")
            self.onefuzz.containers.files.upload_file(crashes, file_path, crash_name)

        report = Report(
            input_blob=BlobRef(
                account=self._get_storage_account(crashes),
                container=crashes,
                name=crash_name,
            ),
            executable=task.config.task.target_exe,
            crash_type="fake crash report",
            crash_site="fake crash site",
            call_stack=["#0 fake", "#1 call", "#2 stack"],
            call_stack_sha256=ZERO_SHA256,
            input_sha256=EMPTY_SHA256,
            asan_log="fake asan log",
            task_id=task_id,
            job_id=task.job_id,
        )

        with tempfile.TemporaryDirectory() as tempdir:
            file_path = os.path.join(tempdir, "report.json")
            with open(file_path, "w") as handle:
                handle.write(report.json())

            self.onefuzz.containers.files.upload_file(
                reports, file_path, crash_name + ".json"
            )


class Debug(Command):
    """ Debug running jobs """

    def __init__(self, onefuzz: Any, logger: logging.Logger):
        super().__init__(onefuzz, logger)
        self.scalesets = DebugScaleset(onefuzz, logger)
        self.repro = DebugRepro(onefuzz, logger)
        self.job = DebugJob(onefuzz, logger)
        self.notification = DebugNotification(onefuzz, logger)
        self.task = DebugTask(onefuzz, logger)



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import configparser
import glob
import os
import subprocess  # nosec
from typing import Dict, List, Optional, Tuple

from onefuzztypes.enums import OS, ContainerType
from onefuzztypes.models import NotificationConfig
from onefuzztypes.primitives import File

from onefuzz.api import Command
from onefuzz.backend import container_file_path

from . import JobHelper

VM_SKU = "Standard_D2s_v3"
VM_COUNT = 1


class OssFuzz(Command):
    """ OssFuzz style jobs """

    def _containers(self, project: str, build: str, platform: OS) -> Dict[str, str]:
        guid = self.onefuzz.utils.namespaced_guid(
            project, build=build, platform=platform.name
        ).hex

        names = {
            "build": "oss-build-%s" % guid,
            "base": "oss-base-%s" % guid,
        }
        return names

    @classmethod
    def _options(_cls, filename: File) -> Tuple[Dict[str, str], List[str]]:
        target_env: Dict[str, str] = {}
        target_options: List[str] = []

        if not os.path.exists(filename):
            return target_env, target_options

        config = configparser.ConfigParser()
        with open(filename, "r") as handle:
            config.read_file(handle)

            if config.has_section("env"):
                target_env.update({x[0].upper(): x[1] for x in config.items("env")})

            if config.has_section("libfuzzer"):
                for arg, value in config.items("libfuzzer"):
                    if arg == "dict":
                        value = "setup/%s" % value
                    target_options.append("-%s=%s" % (arg, value))

        return target_env, target_options

    @classmethod
    def _owners(_cls, filename: File) -> Dict[str, str]:
        tags = {}

        if os.path.exists(filename):
            with open(filename, "r") as handle:
                owners = handle.read()
            owner, ado_path = owners.split(",")
            tags["EMAIL"] = owner
            tags["ADO_PATH"] = ado_path

        return tags

    def _copy_all(self, src_sas: str, dst_sas: str) -> None:
        cmd = [
            "azcopy",
            "sync",
            src_sas,
            dst_sas,
        ]
        self.logger.info("copying base setup")
        subprocess.check_output(cmd)

    def _copy_exe(self, src_sas: str, dst_sas: str, target_exe: File) -> None:
        files: List[File] = [target_exe]
        pdb_path = os.path.splitext(target_exe)[0] + ".pdb"
        if os.path.exists(str(pdb_path)):
            files.append(File(pdb_path))

        for path in files:
            filename = os.path.basename(path)
            src_url = container_file_path(src_sas, filename)
            dst_url = container_file_path(dst_sas, filename)

            cmd = [
                "azcopy",
                "copy",
                src_url,
                dst_url,
            ]
            self.logger.info("uploading %s", path)
            subprocess.check_output(cmd)

    def libfuzzer(
        self,
        project: str,
        build: str,
        pool_name: str,
        duration: int = 24,
        tags: Optional[Dict[str, str]] = None,
        dryrun: bool = False,
        max_target_count: int = 20,
        sync_inputs: bool = False,
        notification_config: Optional[NotificationConfig] = None,
    ) -> None:
        """ OssFuzz style libfuzzer jobs """

        fuzzers = sorted(glob.glob("*fuzzer"))
        if fuzzers:
            platform = OS.linux
        else:
            platform = OS.windows
            fuzzers = sorted(glob.glob("*fuzzer.exe"))

        if dryrun:
            return

        containers = self._containers(project, build, platform)
        container_sas = {}
        for name in containers:
            self.logger.info("creating container: %s", name)
            sas_url = self.onefuzz.containers.create(containers[name]).sas_url
            container_sas[name] = sas_url

        self.logger.info("uploading build artifacts")
        subprocess.check_output(
            [
                "azcopy",
                "sync",
                ".",
                container_sas["build"],
                "--exclude-pattern",
                "*fuzzer_seed_corpus.zip",
            ]
        )
        subprocess.check_output(
            [
                "azcopy",
                "sync",
                ".",
                container_sas["base"],
                '--include-pattern="*.so;*.dll;*.sh;*.ps1',
            ]
        )

        if max_target_count:
            fuzzers = fuzzers[:max_target_count]

        base_helper = JobHelper(
            self.onefuzz,
            self.logger,
            project,
            build,
            "base",
            duration,
            pool_name=pool_name,
            target_exe=File(fuzzers[0]),
        )
        base_helper.platform = platform

        helpers = []
        for fuzzer in [File(x) for x in fuzzers]:
            fuzzer_name = fuzzer.replace(".exe", "").replace("_fuzzer", "")
            self.logger.info("creating tasks for %s", fuzzer)
            self.onefuzz.template.libfuzzer._check_is_libfuzzer(fuzzer)
            helper = JobHelper(
                self.onefuzz,
                self.logger,
                project,
                fuzzer_name,
                build,
                duration,
                job=base_helper.job,
                pool_name=pool_name,
                target_exe=fuzzer,
            )
            helper.platform = platform
            helper.add_tags(tags)
            helper.platform = base_helper.platform
            helper.job = base_helper.job
            helper.define_containers(
                ContainerType.setup,
                ContainerType.inputs,
                ContainerType.crashes,
                ContainerType.reports,
                ContainerType.unique_reports,
                ContainerType.no_repro,
                ContainerType.coverage,
            )
            helper.create_containers()
            helper.setup_notifications(notification_config)

            dst_sas = self.onefuzz.containers.get(
                helper.containers[ContainerType.setup]
            ).sas_url
            self._copy_exe(container_sas["build"], dst_sas, File(fuzzer))
            self._copy_all(container_sas["base"], dst_sas)

            zip_name = fuzzer.replace(".exe", "").replace("_fuzzer", "_seed_corpus.zip")

            if os.path.exists(zip_name) and sync_inputs:
                self.logger.info("uploading seeds")
                helper.upload_inputs_zip(File(zip_name))

            owners_path = File("%s.msowners" % fuzzer.replace(".exe", ""))
            options_path = File("%s.options" % fuzzer.replace(".exe", ""))

            target_env, target_options = self._options(options_path)
            helper.add_tags(self._owners(owners_path))

            # All fuzzers are copied to the setup container root.
            #
            # Cast because `glob()` returns `str`.
            fuzzer_blob_name = helper.target_exe_blob_name(fuzzer, None)

            self.onefuzz.template.libfuzzer._create_tasks(
                job=base_helper.job,
                containers=helper.containers,
                pool_name=pool_name,
                target_exe=fuzzer_blob_name,
                vm_count=VM_COUNT,
                duration=duration,
                target_options=target_options,
                target_env=target_env,
                tags=helper.tags,
            )
            helpers.append(helper)
        base_helper.wait()



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import logging
from typing import Any, Callable, Optional, cast

from signalrcore.hub_connection_builder import HubConnectionBuilder


class Stream:
    connected = None
    hub: Optional[HubConnectionBuilder] = None

    def __init__(self, onefuzz: Any, logger: logging.Logger) -> None:
        self.onefuzz = onefuzz
        self.logger = logger

    def get_token(self) -> str:
        negotiate = self.onefuzz._backend.request("POST", "negotiate")
        return cast(str, negotiate["accessToken"])

    def setup(self, handler: Callable) -> None:
        config = self.onefuzz._backend.request("POST", "negotiate")
        url = config["url"].replace("https://", "wss://")
        self.hub = (
            HubConnectionBuilder()
            .with_url(
                url,
                options={
                    "access_token_factory": self.get_token,
                    "skip_negotiation": True,
                },
            )
            .with_automatic_reconnect(
                {"type": "interval", "keep_alive_interval": 10, "reconnect_interval": 5}
            )
            .build()
        )
        self.hub.on_open(self.on_connect)
        self.hub.on_close(self.on_close)
        self.hub.on("dashboard", handler)
        self.logger.info("connecting to signalr")
        self.hub.start()

    def on_connect(self) -> None:
        self.logger.debug("connected")
        self.connected = True

    def on_close(self) -> None:
        self.logger.debug("disconnected")
        self.connected = False

    def stop(self) -> None:
        if self.hub:
            self.hub.stop()



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from ..api import Command
from .raw import raw
from .top import Top


class Status(Command):
    """ Monitor status of Onefuzz Instance """

    def raw(self) -> None:
        """ Raw status update stream """
        raw(self.onefuzz, self.logger)

    def top(self, show_details: bool = False) -> None:
        """ Onefuzz Top """
        top = Top(self.onefuzz, self.logger, show_details)
        top.run()



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import logging
from datetime import datetime, timedelta, timezone
from enum import Enum
from typing import Any, Dict, List, Optional, Set, Tuple, Union
from uuid import UUID

from onefuzztypes.enums import ContainerType, JobState, NodeState, PoolState, TaskState
from onefuzztypes.models import Job, Node, Pool, Task

MESSAGE = Tuple[datetime, str, str]

MINUTES = 60
HOURS = 60 * MINUTES
DAYS = 24 * HOURS


def fmt_delta(data: timedelta) -> str:
    result = []

    seconds = data.total_seconds()
    for letter, size in [
        ("d", DAYS),
        ("h", HOURS),
        ("m", MINUTES),
    ]:
        part, seconds = divmod(seconds, size)
        if part:
            result.append("%d%s" % (part, letter))

    return "".join(result)


def fmt(data: Any) -> Any:
    if isinstance(data, int):
        return str(data)
    if isinstance(data, str):
        return data
    if isinstance(data, UUID):
        return str(data)[:8]
    if isinstance(data, list):
        return [fmt(x) for x in data]
    if isinstance(data, datetime):
        return data.strftime("%H:%M:%S")
    if isinstance(data, timedelta):
        return fmt_delta(data)
    if isinstance(data, tuple):
        return tuple([fmt(x) for x in data])
    if isinstance(data, Enum):
        return data.name
    if isinstance(data, dict):
        return " ".join(
            sorted(
                [
                    "{}:{}".format(fmt(x).title().replace("_", " "), fmt(y))
                    for (x, y) in data.items()
                ]
            )
        )
    raise NotImplementedError(type(data))


class TopCache:
    JOB_FIELDS = (["Updated", "State", "Job", "Name", "Files"], [1, 1, 1, 1, "100%"])
    TASK_FIELDS = (
        [
            "Updated",
            "State",
            "Job",
            "Task",
            "Type",
            "Name",
            "Files",
            "Pool",
            "Time left",
        ],
        [1, 1, 1, 1, 1, 1, 1, 1, "100%"],
    )
    POOL_FIELDS = (
        ["Updated", "Pool", "Name", "OS", "State", "Nodes"],
        [1, 1, 1, 1, 1, "100%"],
    )

    def __init__(self, onefuzz: "Onefuzz"):
        self.onefuzz = onefuzz
        self.tasks: Dict[UUID, Tuple[datetime, Task]] = {}
        self.jobs: Dict[UUID, Tuple[datetime, Job]] = {}
        self.files: Dict[str, Tuple[Optional[datetime], Set[str]]] = {}
        self.pools: Dict[str, Tuple[datetime, Pool]] = {}
        self.nodes: Dict[UUID, Tuple[datetime, Node]] = {}

        self.messages: List[MESSAGE] = []
        self.endpoint: str = onefuzz._backend.config["endpoint"]
        self.last_update = datetime.now()

    def add_container(self, name: str, ignore_date: bool = False) -> None:
        if name in self.files:
            return
        try:
            files = self.onefuzz.containers.files.list(name)
        except Exception:
            return

        self.add_files(name, set(files.files), ignore_date=ignore_date)

    def add_message(self, name: str, message: Dict[str, Any]) -> None:
        self.last_update = datetime.now()
        data: Dict[str, Union[int, str]] = {}
        for (k, v) in message.items():
            if k in ["task_id", "job_id", "pool_id", "scaleset_id"]:
                k = k.replace("_id", "")
                data[k] = str(v)[:8]
            elif isinstance(v, (int, str)):
                data[k] = v

        as_str = fmt(data)
        messages = [x for x in self.messages if (x[1:] != [name, as_str])][-99:]
        messages += [(datetime.now(), name, as_str)]

        self.messages = messages

    def add_files(
        self, container: str, new_files: Set[str], ignore_date: bool = False
    ) -> None:
        current_date: Optional[datetime] = None
        if container in self.files:
            (current_date, files) = self.files[container]
        else:
            files = set()
        files.update(new_files)
        if not ignore_date:
            current_date = datetime.now()
        self.files[container] = (current_date, files)

    def add_node(
        self, machine_id: UUID, state: NodeState, node: Optional[Node] = None
    ) -> None:
        if state in [NodeState.halt]:
            if machine_id in self.nodes:
                del self.nodes[machine_id]
            return

        if machine_id in self.nodes:
            (_, node) = self.nodes[machine_id]
            node.state = state
            self.nodes[machine_id] = (datetime.now(), node)
        else:
            try:
                if not node:
                    node = self.onefuzz.nodes.get(machine_id)
                self.nodes[node.machine_id] = (datetime.now(), node)
            except Exception:
                logging.debug("unable to find pool: %s", machine_id)

    def add_pool(
        self, pool_name: str, state: PoolState, pool: Optional[Pool] = None
    ) -> None:
        if state in [PoolState.halt]:
            if pool_name in self.pools:
                del self.pools[pool_name]
            return

        if pool_name in self.pools:
            (_, pool) = self.pools[pool_name]
            pool.state = state
            self.pools[pool_name] = (datetime.now(), pool)
        else:
            try:
                if not pool:
                    pool = self.onefuzz.pools.get(pool_name)
                self.pools[pool.name] = (datetime.now(), pool)
            except Exception:
                logging.debug("unable to find pool: %s", pool_name)

    def render_pools(self) -> List:
        results = []

        for (timestamp, pool) in sorted(self.pools.values(), key=lambda x: x[0]):
            timestamps = [timestamp]
            nodes = {}
            for (node_ts, node) in self.nodes.values():
                if node.pool_name != pool.name:
                    continue
                if node.state not in nodes:
                    nodes[node.state] = 0
                nodes[node.state] += 1
                timestamps.append(node_ts)

            timestamps = [timestamp]
            entry = [
                max(timestamps),
                pool.pool_id,
                pool.name,
                pool.os,
                pool.state,
                nodes or "None",
            ]
            results.append(entry)
        return results

    def add_task(
        self,
        task_id: UUID,
        state: TaskState,
        add_files: bool = True,
        task: Optional[Task] = None,
    ) -> None:
        if state in [TaskState.stopping, TaskState.stopped]:
            if task_id in self.tasks:
                del self.tasks[task_id]
            return

        if task_id in self.tasks and self.tasks[task_id][1].state != state:
            (_, task) = self.tasks[task_id]
            task.state = state
            self.tasks[task_id] = (datetime.now(), task)
        else:
            try:
                if task is None:
                    task = self.onefuzz.tasks.get(task_id)
                self.tasks[task.task_id] = (datetime.now(), task)
                if add_files:
                    for container in task.config.containers:
                        self.add_container(container.name)
            except Exception:
                logging.debug("unable to find task: %s", task_id)

    def render_tasks(self) -> List:
        results = []
        for (timestamp, task) in sorted(self.tasks.values(), key=lambda x: x[0]):
            timestamps, files = self.get_file_counts([task])
            timestamps += [timestamp]

            end: Union[str, timedelta] = ""
            if task.end_time:
                end = task.end_time - datetime.now().astimezone(timezone.utc)

            entry = [
                max(timestamps),
                task.state.name,
                task.job_id,
                task.task_id,
                task.config.task.type.name,
                task.config.task.target_exe.replace("setup/", "", 0),
                files,
                task.config.pool.pool_name if task.config.pool else "",
                end,
            ]
            results.append(entry)
        return results

    def add_job(self, job_id: UUID, state: JobState, job: Optional[Job] = None) -> None:
        if state in [JobState.stopping, JobState.stopped]:
            if job_id in self.jobs:
                del self.jobs[job_id]
            return

        if job_id in self.jobs:
            (_, job) = self.jobs[job_id]
            job.state = state
            self.jobs[job_id] = (datetime.now(), job)
        else:
            try:
                if not job:
                    job = self.onefuzz.jobs.get(job_id)
                self.jobs[job_id] = (datetime.now(), job)
            except Exception:
                logging.debug("unable to find job: %s", job_id)

    def render_jobs(self) -> List[Tuple]:
        results: List[Tuple] = []

        for (timestamp, job) in sorted(self.jobs.values(), key=lambda x: x[0]):
            timestamps, files = self.get_file_counts(
                self.get_tasks(job.job_id), merge_inputs=True
            )
            timestamps += [timestamp]

            for to_remove in [ContainerType.coverage, ContainerType.setup]:
                if to_remove in files:
                    del files[to_remove]

            entry = (
                max(timestamps),
                job.state.name,
                job.job_id,
                "%s:%s:%s" % (job.config.project, job.config.name, job.config.build),
                files,
            )
            results.append(entry)

        return results

    def get_file_counts(
        self, tasks: List[Task], merge_inputs: bool = False
    ) -> Tuple[List[datetime], Dict[ContainerType, int]]:
        timestamps = []
        results: Dict[ContainerType, Dict[str, int]] = {}
        for task in tasks:
            for container in task.config.containers:
                if container.name not in self.files:
                    continue
                if merge_inputs and container.type == ContainerType.readonly_inputs:
                    container.type = ContainerType.inputs
                if container.type not in results:
                    results[container.type] = {}
                results[container.type][container.name] = len(
                    self.files[container.name][1]
                )
                container_date = self.files[container.name][0]
                if container_date is not None:
                    timestamps.append(container_date)
        results_merged = {}
        for k, v in results.items():
            value = sum(v.values())
            if value:
                results_merged[k] = value

        return (timestamps, results_merged)

    def get_tasks(self, job_id: UUID) -> List[Task]:
        result = []
        for (_, task) in self.tasks.values():
            if task.job_id == job_id:
                result.append(task)
        return result


from onefuzz.api import Onefuzz  # noqa: E402



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import json
import os
import tempfile
import zipfile
from typing import Any, Dict, List, Optional, Tuple

from onefuzztypes.enums import OS, ContainerType, TaskState
from onefuzztypes.models import Job, NotificationConfig
from onefuzztypes.primitives import Container, Directory, File

from onefuzz.backend import wait

ELF_MAGIC = b"\x7fELF"
DEFAULT_LINUX_IMAGE = "Canonical:UbuntuServer:18.04-LTS:latest"
DEFAULT_WINDOWS_IMAGE = "MicrosoftWindowsDesktop:Windows-10:rs5-pro:latest"


class StoppedEarly(Exception):
    pass


class JobHelper:
    def __init__(
        self,
        onefuzz: "Onefuzz",
        logger: Any,
        project: str,
        name: str,
        build: str,
        duration: int,
        *,
        pool_name: Optional[str] = None,
        target_exe: File,
        platform: Optional[OS] = None,
        job: Optional[Job] = None,
    ):
        self.onefuzz = onefuzz
        self.logger = logger
        self.project = project
        self.name = name
        self.build = build
        self.to_monitor: Dict[str, int] = {}

        if platform is None:
            self.platform = JobHelper.get_platform(target_exe)
        else:
            self.platform = platform

        if pool_name:
            pool = self.onefuzz.pools.get(pool_name)
            if pool.os != self.platform:
                raise Exception(
                    "mismatched os. pool: %s target: %s"
                    % (pool.os.name, self.platform.name)
                )

        self.wait_for_running: bool = False
        self.containers: Dict[ContainerType, Container] = {}
        self.tags: Dict[str, str] = {"project": project, "name": name, "build": build}
        if job is None:
            self.onefuzz.versions.check()
            self.logger.info("creating job (runtime: %d hours)", duration)
            self.job = self.onefuzz.jobs.create(
                self.project, self.name, self.build, duration=duration
            )
            self.logger.info("created job: %s" % self.job.job_id)
        else:
            self.job = job

    def define_containers(self, *types: ContainerType) -> None:
        """ "
        Define default container set based on provided types

        NOTE: in complex scenarios, containers could be defined elsewhere
        """

        for container_type in types:
            if container_type == ContainerType.setup:
                guid = self.onefuzz.utils.namespaced_guid(
                    self.project,
                    self.name,
                    build=self.build,
                    platform=self.platform.name,
                )
            else:
                guid = self.onefuzz.utils.namespaced_guid(self.project, self.name)

            self.containers[container_type] = Container(
                "oft-%s-%s"
                % (
                    container_type.name.replace("_", "-"),
                    guid.hex,
                )
            )

    def create_containers(self) -> None:
        all_containers = [x.name for x in self.onefuzz.containers.list()]
        for (container_type, container_name) in self.containers.items():
            if container_name in all_containers:
                self.logger.info("using container: %s", container_name)
            else:
                self.logger.info("creating container: %s", container_name)
                self.onefuzz.containers.create(
                    container_name, metadata={"container_type": container_type.name}
                )

    def setup_notifications(self, config: Optional[NotificationConfig]) -> None:
        if not config:
            return

        container: Optional[str] = None
        if ContainerType.unique_reports in self.containers:
            container = self.containers[ContainerType.unique_reports]
        else:
            container = self.containers[ContainerType.reports]

        if not container:
            return

        config_dict = json.loads(config.json())
        for entry in self.onefuzz.notifications.list():
            if entry.container == container and entry.config == config_dict:
                self.logger.debug(
                    "notification already exists: %s", entry.notification_id
                )
                return

        self.logger.info("creating notification config for %s", container)
        self.onefuzz.notifications.create(container, config)

    def upload_setup(
        self,
        setup_dir: Optional[Directory],
        target_exe: File,
        setup_files: Optional[List[File]] = None,
    ) -> None:
        if setup_dir:
            target_exe_in_setup_dir = os.path.abspath(target_exe).startswith(
                os.path.abspath(setup_dir)
            )
            if not target_exe_in_setup_dir:
                raise Exception(
                    "target exe `%s` does not occur in setup dir `%s`"
                    % (target_exe, setup_dir)
                )

            self.logger.info("uploading setup dir `%s`" % setup_dir)
            self.onefuzz.containers.files.upload_dir(
                self.containers[ContainerType.setup], setup_dir
            )
        else:
            self.logger.info("uploading target exe `%s`" % target_exe)
            self.onefuzz.containers.files.upload_file(
                self.containers[ContainerType.setup], target_exe
            )

            pdb_path = os.path.splitext(target_exe)[0] + ".pdb"
            if os.path.exists(pdb_path):
                pdb_name = os.path.basename(pdb_path)
                self.onefuzz.containers.files.upload_file(
                    self.containers[ContainerType.setup], pdb_path, pdb_name
                )
        if setup_files:
            for filename in setup_files:
                self.logger.info("uploading %s", filename)
                self.onefuzz.containers.files.upload_file(
                    self.containers[ContainerType.setup], filename
                )

    def upload_inputs(self, path: Directory, read_only: bool = False) -> None:
        self.logger.info("uploading inputs: `%s`" % path)
        container_type = ContainerType.inputs
        if read_only:
            container_type = ContainerType.readonly_inputs
        self.onefuzz.containers.files.upload_dir(self.containers[container_type], path)

    def upload_inputs_zip(self, path: File) -> None:
        with tempfile.TemporaryDirectory() as tmp_dir:
            with zipfile.ZipFile(path, "r") as zip_ref:
                zip_ref.extractall(tmp_dir)

            self.logger.info("uploading inputs from zip: `%s`" % path)
            self.onefuzz.containers.files.upload_dir(
                self.containers[ContainerType.inputs], tmp_dir
            )

    @classmethod
    def get_image(_cls, platform: OS) -> str:
        if platform == OS.linux:
            return DEFAULT_LINUX_IMAGE
        else:
            return DEFAULT_WINDOWS_IMAGE

    @classmethod
    def get_platform(_cls, target_exe: File) -> OS:
        with open(target_exe, "rb") as handle:
            header = handle.read(4)
        if header == ELF_MAGIC:
            return OS.linux
        else:
            return OS.windows

    def wait_on(
        self, wait_for_files: Optional[List[ContainerType]], wait_for_running: bool
    ) -> None:
        if wait_for_files is None:
            wait_for_files = []

        self.to_monitor = {
            self.containers[x]: len(
                self.onefuzz.containers.files.list(self.containers[x]).files
            )
            for x in wait_for_files
        }
        self.wait_for_running = wait_for_running

    def check_current_job(self) -> Job:
        job = self.onefuzz.jobs.get(self.job.job_id)
        if job.state in ["stopped", "stopping"]:
            raise StoppedEarly("job unexpectedly stopped early")

        errors = []
        for task in self.onefuzz.tasks.list(job_id=self.job.job_id):
            if task.state in ["stopped", "stopping"]:
                if task.error:
                    errors.append("%s: %s" % (task.config.task.type, task.error))
                else:
                    errors.append("%s" % task.config.task.type)

        if errors:
            raise StoppedEarly("tasks stopped unexpectedly.\n%s" % "\n".join(errors))
        return job

    def get_waiting(self) -> List[str]:
        tasks = self.onefuzz.tasks.list(job_id=self.job.job_id)
        waiting = [
            "%s:%s" % (x.config.task.type.name, x.state.name)
            for x in tasks
            if x.state not in TaskState.has_started()
        ]
        return waiting

    def is_running(self) -> Tuple[bool, str, Any]:
        waiting = self.get_waiting()
        return (not waiting, "waiting on: %s" % ", ".join(sorted(waiting)), None)

    def has_files(self) -> Tuple[bool, str, Any]:
        self.check_current_job()

        new = {
            x: len(self.onefuzz.containers.files.list(x).files)
            for x in self.to_monitor.keys()
        }

        for container in new:
            if new[container] > self.to_monitor[container]:
                del self.to_monitor[container]
        return (
            not self.to_monitor,
            "waiting for new files: %s" % ", ".join(self.to_monitor.keys()),
            None,
        )

    def wait(self) -> None:
        if self.wait_for_running:
            wait(self.is_running)
            self.logger.info("tasks started")

        if self.to_monitor:
            wait(self.has_files)
            self.logger.info("new files found")

    def target_exe_blob_name(
        self, target_exe: File, setup_dir: Optional[Directory]
    ) -> str:
        # The target executable must end up in the setup container, and the
        # `target_exe` value passed to the tasks must be the name of the target
        # exe _as a blob_ in the setup container.
        if setup_dir:
            # If the user set a `setup_dir`, then `target_exe` must occur inside
            # of it. When we upload the `setup_dir`, the blob name agrees with
            # the `setup_dir`-relative path of the target.
            resolved = os.path.relpath(target_exe, setup_dir)
            if resolved.startswith("..") or resolved == target_exe:
                raise ValueError(
                    "target_exe (%s) is not within the setup directory (%s)"
                    % (target_exe, setup_dir)
                )
            return resolved
        else:
            # If no `setup_dir` was given, we will upload `target_exe` to the
            # root of the setup container created for the user. In that case,
            # the `target_exe` name is just the filename of `target_exe`.
            return os.path.basename(target_exe)

    def add_tags(self, tags: Optional[Dict[str, str]]) -> None:
        if tags:
            self.tags.update(tags)


from onefuzz.api import Onefuzz  # noqa: E402



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from typing import Dict, List, Optional

from onefuzztypes.enums import OS, ContainerType, TaskType
from onefuzztypes.models import Job, NotificationConfig
from onefuzztypes.primitives import Container, Directory, File

from onefuzz.api import Command

from . import JobHelper


class Radamsa(Command):
    """ Pre-defined Radamsa job """

    def basic(
        self,
        project: str,
        name: str,
        build: str,
        *,
        pool_name: str,
        target_exe: File = File("fuzz.exe"),
        setup_dir: Optional[Directory] = None,
        vm_count: int = 2,
        inputs: Optional[Directory] = None,
        reboot_after_setup: bool = False,
        duration: int = 24,
        generator_exe: Optional[str] = None,
        target_options: List[str] = ["{input}"],
        target_env: Optional[Dict[str, str]] = None,
        tags: Optional[Dict[str, str]] = None,
        wait_for_running: bool = False,
        wait_for_files: Optional[List[ContainerType]] = None,
        generator_options: Optional[List[str]] = None,
        radamsa_seed: Optional[int] = None,
        analyzer_exe: Optional[str] = None,
        analyzer_options: Optional[List[str]] = None,
        analyzer_env: Optional[Dict[str, str]] = None,
        existing_inputs: Optional[Container] = None,
        check_asan_log: bool = False,
        check_retry_count: Optional[int] = None,
        disable_check_debugger: bool = False,
        dryrun: bool = False,
        notification_config: Optional[NotificationConfig] = None,
    ) -> Optional[Job]:
        """ Basic radamsa job """

        if inputs is None and existing_inputs is None:
            raise Exception("radamsa requires inputs")

        if dryrun:
            return None

        self.logger.info("creating radamsa from template")

        helper = JobHelper(
            self.onefuzz,
            self.logger,
            project,
            name,
            build,
            duration,
            pool_name=pool_name,
            target_exe=target_exe,
        )

        helper.add_tags(tags)
        helper.define_containers(
            ContainerType.setup,
            ContainerType.crashes,
            ContainerType.reports,
            ContainerType.unique_reports,
            ContainerType.no_repro,
            ContainerType.analysis,
        )
        if existing_inputs:
            self.onefuzz.containers.get(existing_inputs)
            helper.containers[ContainerType.readonly_inputs] = existing_inputs
        else:
            helper.define_containers(ContainerType.readonly_inputs)
        helper.create_containers()
        helper.setup_notifications(notification_config)

        helper.upload_setup(setup_dir, target_exe)
        if inputs:
            helper.upload_inputs(inputs, read_only=True)
        helper.wait_on(wait_for_files, wait_for_running)

        if (
            len(
                self.onefuzz.containers.files.list(
                    helper.containers[ContainerType.readonly_inputs]
                ).files
            )
            == 0
        ):
            raise Exception("Radamsa requires at least one input file")

        target_exe_blob_name = helper.target_exe_blob_name(target_exe, setup_dir)

        tools = Container(
            "radamsa-linux" if helper.platform == OS.linux else "radamsa-win64"
        )
        if generator_exe is None:
            generator_exe = (
                "{tools_dir}/radamsa"
                if helper.platform == OS.linux
                else "{tools_dir}\\radamsa.exe"
            )
        rename_output = True
        if generator_options is None:
            generator_options, rename_output = (
                [
                    "-H",
                    "sha256",
                    "-o",
                    "{generated_inputs}/input-%h.%s",
                    "-n",
                    "100",
                    "-r",
                    "{input_corpus}",
                ],
                False,
            )

        if radamsa_seed is not None:
            generator_options += ["--seed", str(radamsa_seed)]

        self.logger.info("creating radamsa task")

        containers = [
            (ContainerType.tools, tools),
            (ContainerType.setup, helper.containers[ContainerType.setup]),
            (ContainerType.crashes, helper.containers[ContainerType.crashes]),
            (
                ContainerType.readonly_inputs,
                helper.containers[ContainerType.readonly_inputs],
            ),
        ]

        fuzzer_task = self.onefuzz.tasks.create(
            helper.job.job_id,
            TaskType.generic_generator,
            target_exe_blob_name,
            containers,
            pool_name=pool_name,
            duration=duration,
            vm_count=vm_count,
            reboot_after_setup=reboot_after_setup,
            target_options=target_options,
            target_env=target_env,
            generator_exe=generator_exe,
            generator_options=generator_options,
            check_asan_log=check_asan_log,
            check_debugger=not disable_check_debugger,
            tags=helper.tags,
            rename_output=rename_output,
        )

        report_containers = [
            (ContainerType.setup, helper.containers[ContainerType.setup]),
            (ContainerType.crashes, helper.containers[ContainerType.crashes]),
            (ContainerType.reports, helper.containers[ContainerType.reports]),
            (
                ContainerType.unique_reports,
                helper.containers[ContainerType.unique_reports],
            ),
            (ContainerType.no_repro, helper.containers[ContainerType.no_repro]),
        ]

        self.logger.info("creating generic_crash_report task")
        self.onefuzz.tasks.create(
            helper.job.job_id,
            TaskType.generic_crash_report,
            target_exe_blob_name,
            report_containers,
            duration=duration,
            vm_count=1,
            pool_name=pool_name,
            reboot_after_setup=reboot_after_setup,
            target_options=target_options,
            target_env=target_env,
            tags=helper.tags,
            check_asan_log=check_asan_log,
            check_debugger=not disable_check_debugger,
            check_retry_count=check_retry_count,
            prereq_tasks=[fuzzer_task.task_id],
        )

        if helper.platform == OS.windows:
            if analyzer_exe is None:
                analyzer_exe = "cdb.exe"
            if analyzer_options is None:
                analyzer_options = [
                    "-c",
                    "!analyze;q",
                    "-logo",
                    "{output_dir}\\{input_file_name_no_ext}.report",
                    "{target_exe}",
                    "{target_options}",
                ]

            self.logger.info("creating custom analysis")

            analysis_containers = [
                (ContainerType.setup, helper.containers[ContainerType.setup]),
                (ContainerType.tools, tools),
                (ContainerType.analysis, helper.containers[ContainerType.analysis]),
                (ContainerType.crashes, helper.containers[ContainerType.crashes]),
            ]

            self.onefuzz.tasks.create(
                helper.job.job_id,
                TaskType.generic_analysis,
                target_exe_blob_name,
                analysis_containers,
                duration=duration,
                pool_name=pool_name,
                vm_count=vm_count,
                reboot_after_setup=reboot_after_setup,
                target_options=target_options,
                target_env=target_env,
                analyzer_exe=analyzer_exe,
                analyzer_options=analyzer_options,
                analyzer_env=analyzer_env,
                tags=helper.tags,
                prereq_tasks=[fuzzer_task.task_id],
            )

        self.logger.info("done creating tasks")
        helper.wait()
        return helper.job



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import logging
import time
from queue import PriorityQueue
from threading import Thread
from typing import Any, Optional
from uuid import UUID

from onefuzztypes.enums import JobState, NodeState, PoolState, TaskState

from .cache import TopCache
from .signalr import Stream
from .top_view import render


def background_task(queue: PriorityQueue) -> None:
    while True:
        (priority, entry) = queue.get(block=True)
        if entry is None:
            queue.task_done()
            return

        (cmd, args) = entry
        cmd(*args)
        queue.task_done()


class Top:
    def __init__(
        self,
        onefuzz: "Onefuzz",
        logger: logging.Logger,
        show_details: bool,
    ):
        self.onefuzz = onefuzz
        self.logger = logger
        self.show_details = show_details

        self.cache = TopCache(onefuzz)
        self.queue: PriorityQueue = PriorityQueue()
        self.worker = Thread(target=background_task, args=(self.queue,))
        self.worker.start()

    def add_container(self, name: str) -> None:
        if name in self.cache.files:
            return
        self.queue.put((2, (self.cache.add_container, [name, True])))

    def handler(self, message: Any) -> None:
        handlers = {
            "Node": lambda x: self.cache.add_node(
                UUID(x["machine_id"]), NodeState[x["state"]]
            ),
            "Pool": lambda x: self.cache.add_pool(x["name"], PoolState[x["state"]]),
            "Task": lambda x: self.cache.add_task(
                UUID(x["task_id"]), TaskState[x["state"]]
            ),
            "Job": lambda x: self.cache.add_job(
                UUID(x["job_id"]), JobState[x["state"]]
            ),
            "new_file": lambda x: self.cache.add_files(
                x["container"], set([x["file"]])
            ),
        }
        for event in message:
            if event["type"] in handlers:
                handlers[event["type"]](event["data"])
            self.cache.add_message(event["type"], event["data"])

    def setup(self) -> Stream:
        client = Stream(self.onefuzz, self.logger)
        client.setup(self.handler)

        self.logger.info("getting initial data")

        pools = self.onefuzz.pools.list()
        for pool in pools:
            self.cache.add_pool(pool.name, pool.state, pool=pool)

        nodes = self.onefuzz.nodes.list()
        for node in nodes:
            self.cache.add_node(node.machine_id, node.state, node=node)

        jobs = self.onefuzz.jobs.list()

        for job in jobs:
            self.cache.add_job(job.job_id, job.state, job)
            for task in self.onefuzz.tasks.list(job_id=job.job_id):
                self.cache.add_task(
                    task.task_id, task.state, task=task, add_files=False
                )
                for container in task.config.containers:
                    self.add_container(container.name)

        if client.connected is None:
            self.logger.info("waiting for signalr connection")
            while client.connected is None:
                time.sleep(1)

        return client

    def run(self) -> None:
        try:
            client = self.setup()
        except Exception as err:
            self.queue.put((1, None))
            raise err

        error: Optional[Exception] = None
        try:
            self.logger.info("rendering")
            render(self.cache, self.show_details)
            client.stop()
        except Exception as err:
            error = err
        self.queue.put((1, None))
        if error is not None:
            raise error


from ..api import Onefuzz  # noqa: E402



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import logging
import os
import platform
import subprocess  # nosec
import tempfile
from asyncio.subprocess import PIPE
from contextlib import contextmanager
from typing import Generator, Optional


def get_local_tmp() -> Optional[str]:
    # Use $env:LocalAppData on Windows if it's available, as $env:TEMP
    # can frequently have permissions unacceptable for SSH keys

    if platform.system() != "Windows":
        return None

    local_app_data = os.getenv("LOCALAPPDATA")
    if not local_app_data:
        return None

    local_tmp = os.path.join(local_app_data, "temp")
    if os.path.exists(local_tmp):
        return local_tmp

    return None


@contextmanager
def temp_file(
    filename: str, content: str, *, permissions: Optional[str] = None
) -> Generator:
    with tempfile.TemporaryDirectory(dir=get_local_tmp()) as tmpdir:
        full_path = os.path.join(tmpdir, filename)

        logging.debug("creating file %s", full_path)
        with open(full_path, "w") as handle:
            handle.write(content)

        if permissions is not None and platform.system() != "Windows":
            subprocess.check_call(["chmod", permissions, full_path])

        yield full_path

        logging.debug("cleaning up file %s", full_path)


@contextmanager
def build_ssh_command(
    ip: str,
    private_key: str,
    *,
    proxy: Optional[str] = None,
    port: Optional[int] = None,
    command: Optional[str] = None,
) -> Generator:
    with temp_file("id_rsa", private_key, permissions="600") as ssh_key:
        cmd = [
            "ssh",
            "onefuzz@%s" % ip,
            "-i",
            ssh_key,
            "-o",
            "UserKnownHostsFile=/dev/null",
            "-o",
            "StrictHostKeyChecking=no",
        ]

        if proxy:
            cmd += ["-L", proxy]
        if port:
            cmd += ["-p", str(port)]

        log_level = logging.getLogger("nsv-backend").getEffectiveLevel()
        if log_level <= logging.DEBUG:
            cmd += ["-v"]

        if command:
            cmd += [command]

        yield cmd


@contextmanager
def ssh_connect(
    ip: str,
    private_key: str,
    *,
    proxy: Optional[str] = None,
    call: bool = False,
    port: Optional[int] = None,
    command: Optional[str] = None,
) -> Generator:
    with build_ssh_command(
        ip, private_key, proxy=proxy, port=port, command=command
    ) as cmd:
        logging.info("launching ssh: %s", " ".join(cmd))

        if call:
            yield subprocess.call(cmd)
            return

        with subprocess.Popen(
            cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE, bufsize=0
        ) as ssh:
            yield ssh
            ssh.kill()



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import logging
import time
from typing import Any, List

from six.moves import input  # workaround for static analysis

from .signalr import Stream

SIGNALR_CONNECT_TIMEOUT_SECONDS = 0.1


def log_entry(onefuzz: Any, entries: List[Any]) -> None:
    for entry in entries:
        onefuzz.logger.info("%s", entry)


def raw(onefuzz: Any, logger: logging.Logger) -> None:
    client = Stream(onefuzz, logger)
    client.setup(lambda x: log_entry(onefuzz, x))

    while client.connected is None:
        time.sleep(SIGNALR_CONNECT_TIMEOUT_SECONDS)

    wait_for_exit()
    client.stop()


def wait_for_exit() -> None:
    text = ""
    while "exit" not in text:
        print("type exit to stop the log stream")
        text = input("")



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from typing import Dict, List, Optional

from onefuzztypes.enums import ContainerType, TaskType
from onefuzztypes.models import Job, NotificationConfig
from onefuzztypes.primitives import Container, Directory, File

from onefuzz.api import Command

from . import JobHelper

LIBFUZZER_MAGIC_STRING = b"ERROR: libFuzzer"


class Libfuzzer(Command):
    """ Pre-defined Libfuzzer job """

    def _check_is_libfuzzer(self, target_exe: File) -> None:
        """ Look for a magic string """
        self.logger.debug(
            "checking %s for %s", repr(target_exe), repr(LIBFUZZER_MAGIC_STRING)
        )
        with open(target_exe, "rb") as handle:
            data = handle.read()

        if LIBFUZZER_MAGIC_STRING not in data:
            raise Exception("not a libfuzzer binary: %s" % target_exe)

    def _create_tasks(
        self,
        *,
        job: Job,
        containers: Dict[ContainerType, Container],
        pool_name: str,
        target_exe: str,
        vm_count: int = 2,
        reboot_after_setup: bool = False,
        duration: int = 24,
        target_workers: Optional[int] = None,
        target_options: Optional[List[str]] = None,
        target_env: Optional[Dict[str, str]] = None,
        tags: Optional[Dict[str, str]] = None,
        check_retry_count: Optional[int] = None,
        crash_report_timeout: Optional[int] = None,
    ) -> None:

        fuzzer_containers = [
            (ContainerType.setup, containers[ContainerType.setup]),
            (ContainerType.crashes, containers[ContainerType.crashes]),
            (ContainerType.inputs, containers[ContainerType.inputs]),
        ]
        self.logger.info("creating libfuzzer task")
        fuzzer_task = self.onefuzz.tasks.create(
            job.job_id,
            TaskType.libfuzzer_fuzz,
            target_exe,
            fuzzer_containers,
            pool_name=pool_name,
            reboot_after_setup=reboot_after_setup,
            duration=duration,
            vm_count=vm_count,
            target_options=target_options,
            target_env=target_env,
            target_workers=target_workers,
            tags=tags,
        )

        coverage_containers = [
            (ContainerType.setup, containers[ContainerType.setup]),
            (ContainerType.coverage, containers[ContainerType.coverage]),
            (ContainerType.readonly_inputs, containers[ContainerType.inputs]),
        ]
        self.logger.info("creating libfuzzer_coverage task")
        self.onefuzz.tasks.create(
            job.job_id,
            TaskType.libfuzzer_coverage,
            target_exe,
            coverage_containers,
            pool_name=pool_name,
            duration=duration,
            vm_count=1,
            reboot_after_setup=reboot_after_setup,
            target_options=target_options,
            target_env=target_env,
            tags=tags,
            prereq_tasks=[fuzzer_task.task_id],
        )

        report_containers = [
            (ContainerType.setup, containers[ContainerType.setup]),
            (ContainerType.crashes, containers[ContainerType.crashes]),
            (ContainerType.reports, containers[ContainerType.reports]),
            (ContainerType.unique_reports, containers[ContainerType.unique_reports]),
            (ContainerType.no_repro, containers[ContainerType.no_repro]),
        ]

        self.logger.info("creating libfuzzer_crash_report task")
        self.onefuzz.tasks.create(
            job.job_id,
            TaskType.libfuzzer_crash_report,
            target_exe,
            report_containers,
            pool_name=pool_name,
            duration=duration,
            vm_count=1,
            reboot_after_setup=reboot_after_setup,
            target_options=target_options,
            target_env=target_env,
            tags=tags,
            prereq_tasks=[fuzzer_task.task_id],
            target_timeout=crash_report_timeout,
            check_retry_count=check_retry_count,
        )

    def basic(
        self,
        project: str,
        name: str,
        build: str,
        pool_name: str,
        *,
        target_exe: File = File("fuzz.exe"),
        setup_dir: Optional[Directory] = None,
        vm_count: int = 2,
        inputs: Optional[Directory] = None,
        reboot_after_setup: bool = False,
        duration: int = 24,
        target_workers: Optional[int] = None,
        target_options: Optional[List[str]] = None,
        target_env: Optional[Dict[str, str]] = None,
        check_retry_count: Optional[int] = None,
        crash_report_timeout: Optional[int] = None,
        tags: Optional[Dict[str, str]] = None,
        wait_for_running: bool = False,
        wait_for_files: Optional[List[ContainerType]] = None,
        extra_files: Optional[List[File]] = None,
        existing_inputs: Optional[Container] = None,
        dryrun: bool = False,
        notification_config: Optional[NotificationConfig] = None,
    ) -> Optional[Job]:
        """ Basic libfuzzer job """

        # verify containers exist
        if existing_inputs:
            self.onefuzz.containers.get(existing_inputs)

        if dryrun:
            return None

        self.logger.info("creating libfuzzer from template")

        self._check_is_libfuzzer(target_exe)

        helper = JobHelper(
            self.onefuzz,
            self.logger,
            project,
            name,
            build,
            duration,
            pool_name=pool_name,
            target_exe=target_exe,
        )

        helper.add_tags(tags)
        helper.define_containers(
            ContainerType.setup,
            ContainerType.inputs,
            ContainerType.crashes,
            ContainerType.reports,
            ContainerType.unique_reports,
            ContainerType.no_repro,
            ContainerType.coverage,
        )

        if existing_inputs:
            self.onefuzz.containers.get(existing_inputs)
            helper.containers[ContainerType.inputs] = existing_inputs
        else:
            helper.define_containers(ContainerType.inputs)

        helper.create_containers()
        helper.setup_notifications(notification_config)

        helper.upload_setup(setup_dir, target_exe, extra_files)
        if inputs:
            helper.upload_inputs(inputs)
        helper.wait_on(wait_for_files, wait_for_running)

        target_exe_blob_name = helper.target_exe_blob_name(target_exe, setup_dir)

        self._create_tasks(
            job=helper.job,
            containers=helper.containers,
            pool_name=pool_name,
            target_exe=target_exe_blob_name,
            vm_count=vm_count,
            reboot_after_setup=reboot_after_setup,
            duration=duration,
            target_workers=target_workers,
            target_options=target_options,
            target_env=target_env,
            tags=helper.tags,
            crash_report_timeout=crash_report_timeout,
            check_retry_count=check_retry_count,
        )

        self.logger.info("done creating tasks")
        helper.wait()
        return helper.job



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""
Command line interface to the Onefuzz service
"""

import sys

from onefuzz.__version__ import __version__
from onefuzz.api import Command, Endpoint, Onefuzz
from onefuzz.cli import execute_api


def main() -> int:
    return execute_api(Onefuzz(), [Endpoint, Command], __version__)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import sys
from typing import Any, Dict, cast

import setuptools


def get_version() -> str:
    onefuzz: Dict[str, Any] = {}
    with open("onefuzz/__version__.py") as fh:
        # Exec our own __version__.py to pull out version string
        # without import
        exec(fh.read(), onefuzz)  # nosec
    version = onefuzz["__version__"]
    if "-v" in sys.argv:
        index = sys.argv.index("-v")
        sys.argv.pop(index)
        version += ".dev" + sys.argv.pop(index)
    return cast(str, version)


with open("requirements.txt") as f:
    requirements = f.read().splitlines()

    # remove any installer options (see pydantic example)
    requirements = [x.split(" ")[0] for x in requirements]

setuptools.setup(
    name="onefuzz",
    version=get_version(),
    description="Onefuzz Client Library for Python",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/microsoft/onefuzz/",
    author="Microsoft Corporation",
    author_email="fuzzing@microsoft.com",
    license_file="LICENSE",
    packages=setuptools.find_packages(),
    entry_points={"console_scripts": ["onefuzz = onefuzz.__main__:main"]},
    install_requires=requirements,
    zip_safe=False,
    include_package_data=True,
    package_data={"": ["*.md", "*.txt", "data/licenses.json"]},
)



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import logging
import os
import subprocess  # nosec
import tempfile
from contextlib import contextmanager
from typing import Generator


@contextmanager
def rdp_connect(ip: str, password: str, *, port: int) -> Generator:
    FILENAME = "onefuzz.rdp"
    with tempfile.TemporaryDirectory() as tmpdir:
        rdp_file = os.path.join(tmpdir, FILENAME)

        logging.info("encrypting password")

        script = (
            'ConvertTo-SecureString -AsPlainText -Force "%s" | ConvertFrom-SecureString'
            % password
        )

        encrypted = (
            subprocess.check_output(
                ["powershell.exe", "-ExecutionPolicy", "Unrestricted", script]
            )
            .decode()
            .strip()
        )

        content = [
            "full address:s:%s:%d" % (ip, port),
            "username:s:onefuzz",
            "password 51:b:%s" % encrypted,
            "administrative session:i:1",
        ]

        with open(rdp_file, "w") as handle:
            handle.write("\r\n".join(content))

        previous = os.getcwd()
        os.chdir(tmpdir)
        cmd = ["mstsc.exe", FILENAME]
        logging.info("launching rdp: %s", " ".join(cmd))
        yield subprocess.call(cmd)
        os.chdir(previous)



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import atexit
import contextlib
import glob
import json
import logging
import os
import sys
import time
from enum import Enum
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, TypeVar, cast
from urllib.parse import urlparse, urlunparse
from uuid import UUID

import msal
import requests
from azure.storage.blob import ContainerClient
from pydantic import BaseModel
from tenacity import Future as tenacity_future
from tenacity import Retrying, retry
from tenacity.retry import retry_if_exception_type
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_random

_ACCESSTOKENCACHE_UMASK = 0o077

DEFAULT_CONFIG_PATH = os.path.join("~", ".cache", "onefuzz", "config.json")
DEFAULT_TOKEN_PATH = os.path.join("~", ".cache", "onefuzz", "access_token.json")

LOGGER = logging.getLogger("nsv-backend")


@contextlib.contextmanager
def _temporary_umask(new_umask: int) -> Generator[None, None, None]:
    prev_umask = None
    try:
        prev_umask = os.umask(new_umask)
        yield
    finally:
        if prev_umask is not None:
            os.umask(prev_umask)


class Backend:
    def __init__(
        self,
        config: Optional[Dict[str, str]] = None,
        config_path: Optional[str] = None,
        token_path: Optional[str] = None,
    ):
        self.config_path = os.path.expanduser(config_path or DEFAULT_CONFIG_PATH)
        self.token_path = os.path.expanduser(token_path or DEFAULT_TOKEN_PATH)
        self.config = config or {}
        self.token_cache: Optional[msal.SerializableTokenCache] = None
        self.init_cache()
        self.app: Optional[Any] = None
        self.token_expires = 0
        self.load_config()
        self.session = requests.Session()

        atexit.register(self.save_cache)

    def load_config(self) -> None:
        if os.path.exists(self.config_path):
            with open(self.config_path, "r") as handle:
                self.config.update(json.load(handle))

    def save_config(self) -> None:
        with open(self.config_path, "w") as handle:
            json.dump(self.config, handle)

    def init_cache(self) -> None:
        # Ensure the token_path directory exists
        try:
            dir_name = os.path.dirname(self.token_path)
            with _temporary_umask(_ACCESSTOKENCACHE_UMASK):
                os.makedirs(dir_name)
        except FileExistsError:
            pass

        self.token_cache = msal.SerializableTokenCache()
        if os.path.exists(self.token_path):
            with open(self.token_path, "r") as handle:
                self.token_cache.deserialize(handle.read())

    def save_cache(self) -> None:
        if self.token_cache is None:
            return

        if self.token_path is None:
            return

        with _temporary_umask(_ACCESSTOKENCACHE_UMASK):
            with open(self.token_path, "w") as handle:
                handle.write(self.token_cache.serialize())

    def logout(self) -> None:
        self.app = None
        self.token_cache = None
        if os.path.exists(self.token_path):
            os.unlink(self.token_path)

    def headers(self) -> Dict[str, str]:
        value = {}
        if self.config["client_id"] is not None:
            access_token = self.get_access_token()
            value["Authorization"] = "%s %s" % (
                access_token["token_type"],
                access_token["access_token"],
            )
        return value

    def get_access_token(self) -> Any:
        scopes = [self.config["endpoint"] + "/.default"]

        if "client_secret" in self.config:
            return self.client_secret(scopes)
        return self.device_login(scopes)

    def client_secret(self, scopes: List[str]) -> Any:
        if not self.app:
            self.app = msal.ConfidentialClientApplication(
                self.config["client_id"],
                authority=self.config["authority"],
                client_credential=self.config["client_secret"],
                token_cache=self.token_cache,
            )
        result = self.app.acquire_token_for_client(scopes=scopes)
        if "error" in result:
            raise Exception(
                "error: %s\n'%s'"
                % (result.get("error"), result.get("error_description"))
            )
        return result

    def device_login(self, scopes: List[str]) -> Any:
        if not self.app:
            self.app = msal.PublicClientApplication(
                self.config["client_id"],
                authority=self.config["authority"],
                token_cache=self.token_cache,
            )

        accounts = self.app.get_accounts()
        if accounts:
            access_token = self.app.acquire_token_silent(scopes, account=accounts[0])
            if access_token:
                return access_token

        def check_msal_error(value: Dict[str, Any], expected: List[str]) -> None:
            if "error" in value:
                if "error_description" in value:
                    raise Exception(
                        "error: %s\n%s" % (value["error"], value["error_description"])
                    )

                raise Exception("error: %s" % (value["error"]))
            for entry in expected:
                if entry not in value:
                    raise Exception(
                        "interactive login missing value: %s - %s" % (entry, value)
                    )

        LOGGER.info("Attempting interactive device login")
        print("Please login", flush=True)

        flow = self.app.initiate_device_flow(scopes=scopes)
        check_msal_error(flow, ["user_code", "message"])
        print(flow["message"], flush=True)

        access_token = self.app.acquire_token_by_device_flow(flow)
        check_msal_error(access_token, ["access_token"])

        LOGGER.info("Interactive device authentication succeeded")
        print("Login succeeded", flush=True)
        return access_token

    def request(
        self,
        method: str,
        path: str,
        json_data: Optional[Any] = None,
        params: Optional[Any] = None,
        _retry_on_auth_failure: bool = True,
    ) -> Any:
        if not self.config["endpoint"]:
            raise Exception("endpoint not configured")
        url = self.config["endpoint"] + "/api/" + path
        headers = self.headers()
        json_data = serialize(json_data)

        # 401 errors with IDX10501: Signature validation failed occur
        # on rolling new oauth2 client secrets

        # 404 errors happen when new revisions of the functions code are rolling out
        # to the app service environment.
        # TODO: remove this once swapping deployment are in use

        # 502, 503, and 504 errors are often to Azure App issues.
        retry_codes = [401, 404, 429, 502, 503, 504]

        response = None
        for backoff in range(1, 10):
            try:
                LOGGER.debug("request %s %s %s", method, url, repr(json_data))
                response = self.session.request(
                    method,
                    url,
                    headers=headers,
                    json=json_data,
                    params=params,
                    timeout=(30.0, 30.0),
                )

                if response.status_code not in retry_codes:
                    break

                LOGGER.info("request bad status code: %s", response.status_code)
            except requests.exceptions.ConnectionError as err:
                LOGGER.info("request connection error: %s", err)
            except requests.exceptions.ReadTimeout as err:
                LOGGER.info("request timed out: %s", err)

            time.sleep(1.5 ** backoff)

        if response is None:
            raise Exception("request failed: %s %s" % (method, url))

        if response.status_code / 100 != 2:
            error_text = str(
                response.content, encoding="utf-8", errors="backslashreplace"
            )
            raise Exception(
                "request did not succeed: HTTP %s - %s"
                % (response.status_code, error_text)
            )
        return response.json()


def before_sleep(
    retry_object: Retrying, sleep: float, last_result: tenacity_future
) -> None:
    name = retry_object.fn.__name__ if retry_object.fn else "blob function"

    why = getattr(last_result, "_exception")
    if why:
        LOGGER.warning("%s failed with %s, retrying ...", name, repr(why))
    else:
        LOGGER.warning("%s failed, retrying ...", name)


class ContainerWrapper:
    def __init__(self, container_url: str) -> None:
        self.client = ContainerClient.from_container_url(container_url)

    @retry(
        stop=stop_after_attempt(10),
        wait=wait_random(min=1, max=3),
        retry=retry_if_exception_type(),
        before_sleep=before_sleep,
        reraise=True,
    )
    def upload_file(self, file_path: str, blob_name: str) -> None:
        with open(file_path, "rb") as handle:
            self.client.upload_blob(
                name=blob_name, data=handle, overwrite=True, max_concurrency=10
            )
        return None

    def upload_file_data(self, data: str, blob_name: str) -> None:
        self.client.upload_blob(
            name=blob_name, data=data, overwrite=True, max_concurrency=10
        )

    def upload_dir(self, dir_path: str, recursive: bool = True) -> None:
        for path in glob.glob(os.path.join(dir_path, "**"), recursive=recursive):
            if os.path.isfile(path):
                blob_name = os.path.relpath(path, start=dir_path)
                self.upload_file(path, blob_name)

    @retry(
        stop=stop_after_attempt(10),
        wait=wait_random(min=1, max=3),
        retry=retry_if_exception_type(),
        before_sleep=before_sleep,
        reraise=True,
    )
    def delete_blob(self, blob_name: str) -> None:
        self.client.delete_blob(blob_name)
        return None

    @retry(
        stop=stop_after_attempt(10),
        wait=wait_random(min=1, max=3),
        retry=retry_if_exception_type(),
        before_sleep=before_sleep,
        reraise=True,
    )
    def download_blob(self, blob_name: str) -> bytes:
        return cast(bytes, self.client.download_blob(blob_name).content_as_bytes())

    @retry(
        stop=stop_after_attempt(10),
        wait=wait_random(min=1, max=3),
        retry=retry_if_exception_type(),
        before_sleep=before_sleep,
        reraise=True,
    )
    def list_blobs(self, *, name_starts_with: Optional[str] = None) -> List[str]:
        result = [
            x.name for x in self.client.list_blobs(name_starts_with=name_starts_with)
        ]
        return cast(List[str], result)


def container_file_path(container_url: str, blob_name: str) -> str:
    scheme, netloc, path, params, query, fragment = urlparse(container_url)

    blob_url = urlunparse(
        (scheme, netloc, path + "/" + blob_name, params, query, fragment)
    )

    return blob_url


def serialize(data: Any) -> Any:
    if data is None:
        return data
    if isinstance(data, BaseModel):
        return {serialize(a): serialize(b) for (a, b) in data.dict().items()}
    if isinstance(data, dict):
        return {serialize(a): serialize(b) for (a, b) in data.items()}
    if isinstance(data, list):
        return [serialize(x) for x in data]
    if isinstance(data, tuple):
        return tuple([serialize(x) for x in data])
    if isinstance(data, Enum):
        return data.name
    if isinstance(data, UUID):
        return str(data)
    if isinstance(data, (int, str)):
        return data

    raise Exception("unknown type %s" % type(data))


A = TypeVar("A")


def wait(func: Callable[[], Tuple[bool, str, A]], frequency: float = 1.0) -> A:
    """
    Wait until the provided func returns True

    Provides user feedback via a spinner if stdout is a TTY.
    """

    isatty = sys.stdout.isatty()
    frames = ["-", "\\", "|", "/"]
    waited = False
    last_message = None
    result = None

    try:
        while True:
            result = func()
            if result[0]:
                break
            message = result[1]

            if isatty:
                if last_message:
                    if last_message == message:
                        sys.stdout.write("\b" * (len(last_message) + 2))
                    else:
                        sys.stdout.write("\n")
                sys.stdout.write("%s %s" % (frames[0], message))
                sys.stdout.flush()
            elif last_message != message:
                print(message, flush=True)

            last_message = message
            waited = True
            time.sleep(frequency)
            frames.sort(key=frames[0].__eq__)
    finally:
        if waited and isatty:
            print(flush=True)

    return result[2]



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from typing import Dict, List, Optional

from onefuzztypes.enums import OS, ContainerType, StatsFormat, TaskType
from onefuzztypes.models import Job, NotificationConfig
from onefuzztypes.primitives import Container, Directory, File

from onefuzz.api import Command

from . import JobHelper


class AFL(Command):
    """ Pre-defined AFL job """

    def basic(
        self,
        project: str,
        name: str,
        build: str,
        *,
        pool_name: str,
        target_exe: File = File("fuzz.exe"),
        setup_dir: Optional[Directory] = None,
        vm_count: int = 2,
        inputs: Optional[Directory] = None,
        reboot_after_setup: bool = False,
        duration: int = 24,
        target_options: Optional[List[str]] = None,
        supervisor_exe: str = "{tools_dir}/afl-fuzz",
        supervisor_options: List[str] = [
            "-d",
            "-i",
            "{input_corpus}",
            "-o",
            "{runtime_dir}",
            "--",
            "{target_exe}",
            "{target_options}",
        ],
        supervisor_env: Optional[Dict[str, str]] = None,
        supervisor_input_marker: str = "@@",
        tags: Optional[Dict[str, str]] = None,
        wait_for_running: bool = False,
        wait_for_files: Optional[List[ContainerType]] = None,
        afl_container: Optional[Container] = None,
        existing_inputs: Optional[Container] = None,
        dryrun: bool = False,
        notification_config: Optional[NotificationConfig] = None,
    ) -> Optional[Job]:
        """
        Basic AFL job

        :param Container afl_container: Specify the AFL container to use in the job
        """

        if existing_inputs:
            self.onefuzz.containers.get(existing_inputs)

        if dryrun:
            return None

        self.logger.info("creating afl from template")

        target_options = target_options or ["{input}"]

        helper = JobHelper(
            self.onefuzz,
            self.logger,
            project,
            name,
            build,
            duration,
            pool_name=pool_name,
            target_exe=target_exe,
        )
        helper.add_tags(tags)
        helper.define_containers(
            ContainerType.setup,
            ContainerType.crashes,
            ContainerType.reports,
            ContainerType.unique_reports,
        )
        if existing_inputs:
            self.onefuzz.containers.get(existing_inputs)
            helper.containers[ContainerType.inputs] = existing_inputs
        else:
            helper.define_containers(ContainerType.inputs)

        helper.create_containers()
        helper.setup_notifications(notification_config)
        helper.upload_setup(setup_dir, target_exe)
        if inputs:
            helper.upload_inputs(inputs)
        helper.wait_on(wait_for_files, wait_for_running)

        if (
            len(
                self.onefuzz.containers.files.list(
                    helper.containers[ContainerType.inputs]
                ).files
            )
            == 0
        ):
            raise Exception("AFL requires at least one input")

        target_exe_blob_name = helper.target_exe_blob_name(target_exe, setup_dir)

        if afl_container is None:
            afl_container = Container(
                "afl-linux" if helper.platform == OS.linux else "afl-windows"
            )

        # verify the AFL container exists
        self.onefuzz.containers.get(afl_container)

        containers = [
            (ContainerType.tools, afl_container),
            (ContainerType.setup, helper.containers[ContainerType.setup]),
            (ContainerType.crashes, helper.containers[ContainerType.crashes]),
            (ContainerType.inputs, helper.containers[ContainerType.inputs]),
        ]

        self.logger.info("creating afl fuzz task")
        fuzzer_task = self.onefuzz.tasks.create(
            helper.job.job_id,
            TaskType.generic_supervisor,
            target_exe_blob_name,
            containers,
            pool_name=pool_name,
            duration=duration,
            vm_count=vm_count,
            reboot_after_setup=reboot_after_setup,
            target_options=target_options,
            supervisor_exe=supervisor_exe,
            supervisor_options=supervisor_options,
            supervisor_env=supervisor_env,
            supervisor_input_marker=supervisor_input_marker,
            stats_file="{runtime_dir}/fuzzer_stats",
            stats_format=StatsFormat.AFL,
            task_wait_for_files=ContainerType.inputs,
            tags=helper.tags,
        )

        report_containers = [
            (ContainerType.setup, helper.containers[ContainerType.setup]),
            (ContainerType.crashes, helper.containers[ContainerType.crashes]),
            (ContainerType.reports, helper.containers[ContainerType.reports]),
            (
                ContainerType.unique_reports,
                helper.containers[ContainerType.unique_reports],
            ),
        ]

        self.logger.info("creating generic_crash_report task")
        self.onefuzz.tasks.create(
            helper.job.job_id,
            TaskType.generic_crash_report,
            target_exe_blob_name,
            report_containers,
            pool_name=pool_name,
            duration=duration,
            vm_count=1,
            reboot_after_setup=reboot_after_setup,
            target_options=target_options,
            check_debugger=True,
            tags=tags,
            prereq_tasks=[fuzzer_task.task_id],
        )

        self.logger.info("done creating tasks")
        helper.wait()
        return helper.job



#!/usr/bin/env python
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import logging
from typing import Optional

from onefuzztypes.enums import ContainerType

from .api import Command, Onefuzz
from .templates.afl import AFL
from .templates.libfuzzer import Libfuzzer
from .templates.ossfuzz import OssFuzz
from .templates.radamsa import Radamsa


class Template(Command):
    """ Pre-defined job templates """

    def __init__(self, onefuzz: Onefuzz, logger: logging.Logger) -> None:
        super().__init__(onefuzz, logger)
        self.libfuzzer = Libfuzzer(onefuzz, logger)
        self.afl = AFL(onefuzz, logger)
        self.radamsa = Radamsa(onefuzz, logger)
        self.ossfuzz = OssFuzz(onefuzz, logger)

    def stop(
        self,
        project: str,
        name: str,
        build: Optional[str],
        delete_containers: bool = False,
        stop_notifications: bool = False,
    ) -> None:
        SAFE_TO_REMOVE = [
            ContainerType.crashes,
            ContainerType.setup,
            ContainerType.inputs,
            ContainerType.reports,
            ContainerType.unique_inputs,
            ContainerType.unique_reports,
            ContainerType.no_repro,
            ContainerType.analysis,
            ContainerType.coverage,
            ContainerType.readonly_inputs,
        ]

        msg = ["project:%s" % project, "name:%s" % name]
        if build is not None:
            msg.append("build:%s" % build)
        self.logger.info("stopping %s" % " ".join(msg))

        containers = [x.name for x in self.onefuzz.containers.list()]
        jobs = self.onefuzz.jobs.list()
        for job in jobs:
            if job.config.project != project:
                continue

            if job.config.name != name:
                continue

            if build is not None and job.config.build != build:
                continue

            if job.state not in ["stopped", "stopping"]:
                self.logger.info("stopping job: %s", job.job_id)
                self.onefuzz.jobs.delete(job.job_id)

            tasks = self.onefuzz.tasks.list(job_id=job.job_id)
            for task in tasks:
                if task.state not in ["stopped"]:
                    self.logger.info("stopping task: %s", task.task_id)
                    self.onefuzz.tasks.delete(task.task_id)

                if delete_containers:
                    to_remove = []
                    for container in task.config.containers:
                        if container.type not in SAFE_TO_REMOVE:
                            self.logger.info("not removing: %s", container)
                            continue
                        to_remove.append(container.name)
                    for name in to_remove:
                        if name in containers:
                            self.logger.info("removing container: %s", name)
                            self.onefuzz.containers.delete(name)
                            containers.remove(name)

                if stop_notifications:
                    notifications = self.onefuzz.notifications.list()
                    for container in task.config.containers:
                        for notification in notifications:
                            if container.name == notification.container:
                                self.logger.info(
                                    "removing notification: %s",
                                    notification.notification_id,
                                )
                                self.onefuzz.notifications.delete(
                                    notification.notification_id
                                )


Template.stop.__doc__ = "stop a template job"
